# -*- python -*-
# $RCSfile: bdycondition.py,v $
# $Revision: 1.101 $
# $Author: langer $
# $Date: 2008-06-12 15:34:01 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# A new, leaner, meaner, boundary condition class.  For the
# point-like ones, the iteration over nodes and point evaluation,
# as required, will be done elsewhere.  All we need to be able
# to do here is take a node and a profile result, and apply it.

from ooflib.SWIG.common import ooferror
from ooflib.SWIG.engine import ooferror2
from ooflib.SWIG.common import switchboard
from ooflib.SWIG.common import lock
from ooflib.SWIG.engine import boundarycond
from ooflib.common import debug
from ooflib.common import labeltree
from ooflib.common import registeredclass
from ooflib.common.IO import automatic
from ooflib.common.IO import parameter
from ooflib.common.IO import whoville
from ooflib.common.IO import xmlmenudump
from ooflib.engine import problem
from ooflib.engine import boundary
from ooflib.engine import profile
from ooflib.engine import skeletoncontext
from ooflib.engine.IO import meshparameters
import ooflib.engine.mesh
import string
#Interface branch
from ooflib.engine.IO import materialparameter

# It's useful in the editor if all the BC's have shared
# parameters, to the greatest degree possible.

bc_field_param_group = parameter.ParameterGroup(
    meshparameters.FieldParameter('field', tip=parameter.emptyTipString),
    meshparameters.FieldIndexParameter('field_component',
                                       tip=parameter.emptyTipString))

bc_equation_param_group = parameter.ParameterGroup(
    meshparameters.EquationBCParameter('equation',
                                       tip=parameter.emptyTipString),
    meshparameters.FieldIndexParameter('eqn_component',
                                       tip=parameter.emptyTipString))


# bc_boundary_param = skeletongroupparams.BoundaryParameter(
#     "boundary",tip="Boundary to which this condition applies.")
bc_boundary_param = meshparameters.MeshBoundaryParameter(
    "boundary", tip="Boundary to which this condition applies.")

bc_edge_bdy_param = meshparameters.MeshEdgeBdyParameter(
    "boundary", tip="Edge boundary to which this condition applies.")

bc_periodic_edge_bdy_param = meshparameters.MeshPeriodicEdgeBdyParameter(
    "boundary",
    tip="Pair of periodic edge boundaries to which this condition applies")

bc_point_bdy_param = meshparameters.MeshPointBdyParameter(
    "boundary", tip="Point boundary to which this condition applies.")




# Boundary conditions are constructed in association with a particular
# profile and boundary, and can't really exist without these.
# Boundary conditions cannot be shared between profiles or between
# boundaries.
class BC(registeredclass.RegisteredClass):
    registry = []
    def __init__(self, profile, boundary, user_enable=True, visible=True,
                 subordinate=False):
        self.profile = profile     # Profile object.
        self.boundary = boundary   # Name of the boundary.

        # Unpack the NamedProfile, if you got one.
        self.profile_obj = self.profile.unpack()

        # These variables are set by add_to_mesh.
        self.boundary_obj = None
        self.name = None
        self.mesh_obj = None    # mesh context
        self.mesh = None        # mesh name

        self.sw_cb = [switchboard.requestCallback(
            "new namedprofile obj", self.replace_profile)]

        # BC's start out inactive, but are auto-activated at
        # add-to-mesh time.  If they are ever explicitly inactivated
        # by the user, the explicit_inactive flag gets set, and
        # auto_activate will not make them active.
        self.computableDict = {}
        self.explicit_disable = not user_enable

        # The former "promotion" flag is now the visible flag.
        self.visible = visible
        # Subordinate boundary conditions are ones created by another
        # condition, such as the FloatBCs that are used to implement
        # PeriodicBCs.  It's possible that subordinate==visible.
        self.subordinate = subordinate


    def femesh(self):
        # self.mesh_obj is set, when "add_to_mesh" is called.
        return self.mesh_obj.femesh()

    # Status-setting function -- enables the BC if it's possible to do
    # so and the user has never said not to.  This function should be
    # called whenever the fields/equations/flux activation info in the
    # mesh changes.  Subclasses must have the "computable" function,
    # which answers the "do you make sense" question, independently of
    # whether the BC will actually be enabled.
    def auto_enable(self, subpcontext):
        subproblem = subpcontext.getObject()
        self.computableDict[subproblem] = self.computable(subproblem)

    # Functions for explicit enabling or disabling of BCs by users.

    def toggle_explicit_disable(self):
        msg = 0
        # check for collisions if we are enabling a bc
        if self.explicit_disable:
            msg = self.check(self.mesh)
        if msg:
            raise ooferror.ErrSetupError(msg)
        else:
            self.explicit_disable = not self.explicit_disable
            self.mesh_obj.bdys_changed()
            switchboard.notify("boundary conditions changed",
                               self.mesh_obj, self.name, self.isVisible())

    def explicitly_enable(self):
        if self.is_explicitly_disabled():
            self.toggle_explicit_disable()

    def explicitly_disable(self):
        if not self.is_explicitly_disabled():
            self.toggle_explicit_disable()

    def is_explicitly_disabled(self):
        return self.explicit_disable

    def is_disabled(self, subproblem):
        return self.explicit_disable or not self.computableDict[subproblem]

    # Some BCs will not be visible in the GUI.  For instance if
    # they're the result of promoting other BCs or the auxiliary
    # floatBC's used by periodicBC's
    def isVisible(self):
        return self.visible

    def __eq__(self, other):
        return id(self)==id(other)
    
    # Add_to_mesh is separate from the construction process for the
    # convenience of the boundary-condition editing, which is
    # accomplished by creating a new boundary condition and adding it
    # to the mesh with the name of a recently-removed one. 
    def add_to_mesh(self, name, mesh):
        self.name = name 
        self.mesh = mesh # Mesh name.
        
        self.mesh_obj = ooflib.engine.mesh.meshes[mesh] # Who object.

        try:
            self.boundary_obj = self.mesh_obj.getBoundary(self.boundary)
        except KeyError:
            raise

        self.profile_obj.addCondition(self)
        self.boundary_obj.addCondition(self)

        # Auto-enable must precede the mesh_obj add call below, which
        # builds the display string, which depends on side-effects of
        # auto_enable. (Really?)
        for subproblem in self.mesh_obj.subproblems():
            self.auto_enable(subproblem)
        self.mesh_obj.addBdyCondition(name, self)

        
    def check(self, mesh, exclude=None):
        # Check that this bc doesn't conflict with any other bc on
        # this mesh. This doesn't check the *values* of the boundary
        # conditions.  It only checks that two Dirichlet or floating
        # boundary conditions don't apply to the same degrees of
        # freedom on the same boundary.
        # If the 'exclude' variable is equal to some other bc, then
        # conflicts with that bc aren't counted.

        try:
            mesh_obj = ooflib.engine.mesh.meshes[mesh]
        except:
            return "There is no mesh named '%s'!" % mesh
        try:
            boundary_obj = mesh_obj.getBoundary(self.boundary)
        except:
            return "There is no boundary named '%s'!" % self.boundary
        return boundary_obj.checkCondition(self, exclude=exclude)


    def remove_from_mesh(self):
        self.mesh_obj.rmBdyCondition(self)
        # rmBdyCondition calls disconnect, below.

    def rename(self, newname):
        if self.boundary_obj:
            self.boundary_obj.renameCondition(self.name, newname)
        self.name = newname

    # Called prior to deletion, undo the additions we did on __init__.
    # Because of the circular references, as well as control over the
    # timing, these operations are unsuitable for a __del__ operation.
    def disconnect(self):
        self.profile_obj.removeCondition(self)
        self.boundary_obj.removeCondition(self)
        self.boundary_obj = None
        self.mesh = None
        self.mesh_obj = None
        self.computableDict = {}

        # Break circular references, but leave in the ones
        # corresponding to parameters.
        
        # self.boundary = None
        # self.profile=None
        self.profile_obj = None
        # Clean up callbacks.
        map(switchboard.removeCallback, self.sw_cb)


    # Switchboard callback for "new namedprofile obj". Called when
    # profiles are edited.  The NamedProfile has a new .obj, and the
    # conditions must all update accordingly.
    def replace_profile(self, profile):
        if self.profile == profile:
            self.profile_obj = self.profile.unpack()
            self.profile_obj.addCondition(self)
            self.mesh_obj.bdys_changed()

    # Used by the GUI.
    def display_string(self):
        if self.boundary_obj and self.profile_obj:
            resstr = self.boundary_obj.name() + " / " + self.display() + \
                     " / %s" % self.profile_obj.descriptor()
        else:
            #Interface JumpBC has no boundary_obj (None)
            if isinstance(self,JumpBC) and self.profile_obj:
                resstr = self.boundary + " / " + self.display()
                         #" / %s" % self.profile_obj.descriptor()
            else:
                resstr = "Unassigned / " + self.display()

        # If there's no Mesh, or if the BC isn't computable on any of
        # the Mesh's subproblems, then mark the BC as "disabled".
        if self.mesh_obj:
            for s in self.mesh_obj.subproblems():
                if not self.is_disabled(s.getObject()): 
                    return resstr
                elif self.explicit_disable:
                    disstr = "(user disabled) "
                else:
                    disstr = "(incomputable) " #better description?
        return disstr + resstr

    tip = "Different kinds of boundary conditions."
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/bc.xml')

# Individual boundary condition classes take the field and equation
# data in the form of a component object, which they must then change
# to a separate field and component for actual evaluation.  Boundary
# conditions are also specific to a particular mesh and boundary.
# They also know about a profile -- the object passed in may either be
# an existing profile in the current mesh, or a new unnamed profile.
# In either case, the passed in object is the appropriate subclass of
# ProfileFinder.

class DirichletBC(BC):
    def __init__(self, field, field_component, equation, eqn_component,
                 profile, boundary, 
                 offset=0.0, user_enable=True, visible=True, promoted=False,
                 subordinate=False):
        self.field = field
        self.field_component = field_component
        self.equation = equation
        self.eqn_component = eqn_component

        # Parent class __init__ needs to be last, because it does
        # collision detection on the boundary, which means the field
        # and equation have to already be set.
        BC.__init__(self, profile, boundary, user_enable=user_enable,
                    visible=visible, subordinate=subordinate)

        # When a DirichletBC is auto-generated from a FloatBC, it can
        # acquire an offset.  The visible flag is set by BC__init__  
        self.offset = offset
        self.promoted = promoted


    # In most places, promotion has been superceded by the visible
    # flag.  It is still used in boundary.py for the reset function.
    def promotion(self):
        return self.promoted

    
    # "Copy" makes a new BC with a new name on a new boundary, but
    # with the same other parameters, except explicit-enabling.
    def copy(self, boundary):
        return DirichletBC(self.field, self.field_component,
                           self.equation, self.eqn_component,
                           self.profile, boundary,
                           user_enable=not self.explicit_disable)

    # Dispatch routine to call the right boundary-addition gizmo.
    def addToBoundary(self, boundary):
        boundary.addFixedCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkFixedCondition(self, exclude)

    # Am I computable on the given subproblem of the current mesh?
    def computable(self, subproblem):
        if not self.mesh_obj:
            return False
        return (subproblem.is_defined_field(self.field) and 
                subproblem.is_active_field(self.field) and 
                subproblem.is_active_equation(self.equation))
        
        
    # BC is applied through its "call" method, which just sets
    # the value of the appropriate DOF to the passed-in value, plus
    # the offset, and marks the corresponding equation as
    # dependent.
    def __call__(self, subproblem, node, value):
        if self.is_disabled(subproblem):
            return
        fldcomp = self.field.getIndex(self.field_component).integer()
        try:
            # If the field isn't defined or the equation isn't active
            # at this node (because bcs are applied to Meshes but
            # fields & equations are handled on SubProblems), then
            # either Field.setvalue or SubProblem.map_forward_equation
            # will fail.  In that case, there's nothing to do -- the
            # bc can't be applied.
            self.field.setvalue(node, fldcomp, value+self.offset)
            eqncomp = self.equation.getIndex(self.eqn_component).integer()
            nodalEqn = subproblem.map_forward_equation(self.equation, node,
                                                       eqncomp)
        except ooferror2.ErrNoSuchField:
            return
        dof = self.field.dof(node, fldcomp)
        dof.fix()
        nodalEqn.dependent()

    def intersect(self, other):
        # Other should be another DirichletBC or a FloatBC.
        return (self.field == other.field and
                self.field_component == other.field_component and
                self.equation == other.equation and
                self.eqn_component == other.eqn_component)

    def display(self):
        return "DirichletBC / %s[%s] / %s[%s]" % (`self.field`,
                                              self.field_component,
                                              `self.equation`,
                                              self.eqn_component)
    
registeredclass.Registration(
    "Dirichlet",
    BC, DirichletBC,
    ordering=1,
    params = (bc_field_param_group +
              bc_equation_param_group + [ 
            parameter.RegisteredParameter(
                'profile', reg=profile.Profile, value=None,
                tip='Profile describing how this condition varies.'),
            bc_boundary_param
            ]),
    tip="Fix the values of a Field along a boundary.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/dirichlet.xml')
    ) 


class ForceBC(BC):
    def __init__(self, equation, eqn_component, profile, boundary,
                 user_enable=True):
        self.equation = equation
        self.eqn_component = eqn_component

        BC.__init__(self, profile, boundary,user_enable=user_enable)

    def copy(self, boundary):
        return ForceBC(self.equation, self.eqn_component,
                       self.profile, boundary,
                       user_enable=not self.explicit_disable)
    #
    # def write(self, file, indent):
    #     file.write(indent+"ForceBC("+problem.inversedict[self.equation]+",")
    #     file.write("%d)" % self.equationComponent)
    #
    def addToBoundary(self, boundary):
        boundary.addForceCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkForceCondition(self, exclude)

    def computable(self, subproblem):
        if not self.mesh_obj:
            return None
        return subproblem.is_active_equation(self.equation)
    #
    # Applied via "call" method, just creates an applicator if it
    # doesn't already have one, and calls it with the node and
    # equation component, and the value to set.  
    #
    def __call__(self, subproblem, node, value):
        if self.is_disabled(subproblem):
            return
        eqnindex = self.equation.getIndex(self.eqn_component).integer()
        try:
            boundarycond.applyForceBC(subproblem, self.equation,
                                      node, eqnindex, value)
        except ooferror2.ErrNoSuchField: # eqn not active at that node
            pass
    def display(self):
        return "ForceBC / %s" % `self.equation`



registeredclass.Registration(
    "Generalized Force",
    BC,
    ForceBC,
    ordering=2,
    params = bc_equation_param_group + [
        parameter.RegisteredParameter(
            'profile', reg=profile.Profile, value=None,
            tip='Profile describing how this condition varies.'),
        bc_point_bdy_param],
    tip="Apply a force to all nodes on a boundary.  A force is a divergence of a flux (eg, stress, heat flux).  The condition is applied independently at each node, and is therefore sensitive to node density.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/forcebc.xml')
    )


## Float boundary condition -- similar to a DirichletBC in basic
## structure, but with major differences in the apply function.  What
## 'apply' does is go through the mapping lists passed to the mesh by
## the matrix extraction routine, and set them up so that all the rows
## and columns corresponding to the DOFs and nodalequations of the
## floating boundary condition are remapped to a single row/colum,
## respectively, of the new mapping.

## They also have an unravelling routine which unmaps the DOFs
## after solution.

## The 'profile' function has the same structure but a different
## meaning here -- it's an offset, the true fully-solved value of the
## DOF is the profile value plus another value, such that the sum
## equals the sum arrived at by the solver.


class FloatBC(BC):
    def __init__(self, field, field_component, equation, eqn_component,
                 profile, boundary, user_enable=True, visible=True,
                 subordinate=False):
        self.field = field
        self.field_component = field_component
        self.equation = equation
        self.eqn_component = eqn_component
        BC.__init__(self, profile, boundary, user_enable=user_enable,
                    visible=visible, subordinate=subordinate)
        self.reset()

    def copy(self, boundary):
        return FloatBC(self.field, self.field_component,
                       self.equation, self.eqn_component,
                       self.profile, boundary,
                       user_enable=not self.explicit_disable)

    # This function can be called either by init or from the boundary,
    # when it wants to reset stateful boundary conditions.  This
    # provides a handy way of pretend-encapsulating the state variables.
    # FloatBCs are stateful because intersecting combinations of them
    # need to aggregate their degrees of freedom in a sensible way,
    # with consistent profiles and without double-counting.  This is
    # what the traversal process does.
    def reset(self):
        # Below here are the "stateful" data members.  
        self.applicator = None     # C applicator -- "None" means first-call.
        self.dofmappingIndex = -1  # Set either locally or by parents
        self.eqnmappingIndex = -1  #    on the intersection tree.
        self.expandfirst = None    # Detects first call to "expand"
        self.offset = 0.0          # Local profile offset relative to parent.
        self.profileStart = 0.0    # Master start value, set by root.
        self.profileOffset = 0.0   # Cumulative total offset from intersections
        self.active = 1            # 
        self.dupnodes = {}         # Rationalized duplicate set.
        self.nodeshares = {}       # BC-key'd dict, parent : dict of dupnodes.
        self.intersectors = {}     # BC-key'd dict, child : offset
        self.intersected = None    # Intersection state.
        # States are None, "mapped", and "traversed."


        # A FloatBC is active by default, but can be deactivated
        # if it gets promoted to a dirichlet BC.  This happens when
        # the "fix" method is called.
##        # The floatappset member helps to encode state information -- the
##        # applicator knows its first call, which is why it's important
##        # to de-allocate it at the end.
        #   "Activeness" is checked for locally, so that active and
        # inactive floatBC's present the same interface, but inactive
        # ones don't actually do anything.  The only other place
        # activeness is checked for is in the intersection routine.
        #
        # Duplicate nodes can arise from intersection -- the calls
        # simply do nothing at all in that case.  The first node of
        # the parent of an intersection tree is guaranteed not to
        # be duplicate to the parent, so all the order-depdendent
        # stuff still works.


    # Dispatch routine, called from the boundary, calls back to the
    # correct boundary-addition routine.
    def addToBoundary(self, boundary):
        boundary.addFloatCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkFloatCondition(self, exclude)

    def computable(self, subproblem):
        if not self.mesh_obj:
            return None
        return (subproblem.is_defined_field(self.field) and
                subproblem.is_active_field(self.field) and
                subproblem.is_active_equation(self.equation))
        
    
    def __call__(self, subproblem, node, value):
        if self.is_disabled(subproblem):
            return
##        # Statefulness is in the floatappset list -- if it's null, this is
##        # the first call, you need to set up one for each field component.
        # if a boundary is intersected at each node, the applicator
        # will not be initialized
        if self.active and not self.dupnodes.has_key(node):
            fldcomp = self.field.getIndex(self.field_component).integer()
            #This modification would allow out-of-plane fields (such
            #as the out-of-plane derivatives of the displacement u)
            #for self.field. Since the components of the Plane_Stress
            #equation are specified by tensor indices
            #(self.eqn_component="xz","yz" or "zz"), we cannot use
            #equation.getIndex(self.eqn_component).integer() as it
            #returns the values (4,3,2) (i.e. voight, see problem.py
            #and fieldindex.h)
            if self.eqn_component=='zz':
                eqncomp=2
            elif self.eqn_component=='yz' or self.eqn_component=='zy':
                eqncomp=1
            elif self.eqn_component=='xz' or self.eqn_component=='zx':
                eqncomp=0
            else:
                eqncomp = self.equation.getIndex(self.eqn_component).integer()
                if eqncomp>1:
                    #The exact conjugacy may not be important, so
                    #simply subtracting 2 should also work for the
                    #Plane_Stress case.  (eqncomp becomes 2 for
                    #Plane_Heat_Flux, really need to convert it to 0.
                    #One also gets '' for self.field_component when
                    #self.field is Temperature_z.)
                    eqncomp-=2
            
            value += self.profileOffset
            if self.applicator:
                # This is a non-first call, just do editmap.
                try:
                    self.applicator.editmap(subproblem,
                                            value, node,
                                            self.field, fldcomp,
                                            self.equation, eqncomp,
                                            self.dofmappingIndex,
                                            self.eqnmappingIndex,
                                            self.profileStart)
                except ooferror2.ErrNoSuchField:
                    # Field or Eqn isn't defined at the node
                    pass
            else:
                # This is the first call to this particular FloatBC.
                applicator = boundarycond.FloatBCApp()
                if not self.intersected:
                    # We are either the root of an intersection tree,
                    # or unintersected -- find out what mapping targets to use.
                    try:
                        self.dofmappingIndex = applicator.getdofindex(
                            subproblem, node, self.field, fldcomp)
                        self.eqnmappingIndex = applicator.geteqnindex(
                            subproblem, node, self.equation, eqncomp)
                    except ooferror2.ErrNoSuchField:
                        pass
                    else:
                        self.profileStart = value
                        self.applicator = applicator

                        # Broadcast these to the intersecting set.
                        for intersector in self.intersectors.keys():
                            intersector.propagate_(self.dofmappingIndex,
                                                   self.eqnmappingIndex,
                                                   self.profileStart)
                else:
                    # If we're here, it means this is the first call
                    # to a non-root member of an intersection tree,
                    # so we need to call editmap on the first node,
                    try:
                        self.applicator = applicator
                        self.applicator.editmap(
                            subproblem, value, node,
                            self.field, fldcomp,
                            self.equation, eqncomp,
                            self.dofmappingIndex,
                            self.eqnmappingIndex,
                            self.profileStart)
                    except ooferror2.ErrNoSuchField:
                        pass

    # This is the function that handles the propagation of parameters
    # through the intersections -- recursively propagates everything.
    def propagate_(self, dofmappingIndex, eqnmappingIndex, profileStart):
        if self.intersected != 'traversed':
            raise ooferror.ErrPyProgrammingError(
                "Data propagated to incomplete Float.")
        self.dofmappingIndex = dofmappingIndex
        self.eqnmappingIndex = eqnmappingIndex
        self.profileStart = profileStart
        for intersector in self.intersectors.keys():
            intersector.propagate_(dofmappingIndex, eqnmappingIndex,
                                  profileStart)
            
    # "is_disabled" indicates whether the fields and equations are
    # defined, or whether or not the BC has been turned off by the
    # user.  "Active" refers to whether or not this BC has survived
    # the promotion process.  They are orthogonal.
    def contrib_rhs(self, subproblem):
        if self.is_disabled(subproblem):
            return
        if self.active and self.applicator:
            # The only case where a bc should be active but without an
            # applicator is when a bc is a leaf in the boundary
            # condition intersection tree and all of the nodes are
            # intersections with other floating bc's.  This is the
            # case for one of the four auxiliary corner point bc's
            # that are created when periodic bc's are enforced in both
            # directions. As long as all intersecting floating
            # boundary conditions are consistent, it is appropriate to
            # ignore this case.  Consistency is checked for in an
            # earlier step in the function floatBCResolve which is
            # called by precompute.
            self.applicator.profile_rhs(subproblem)
            self.applicator = None  # Null this, anticipating next round.
            self.expandfirst = None


    # After matrix solution, need to fill in the DOF values
    # according to the profile function.  Tuple length-matches not
    # checked, because if they're wrong, we failed in "invoke."
    # This routine breaks if the nodes don't come in the
    # same order as they did in invoke.  (More specifically, the same
    # node must occur first.)

    def expand(self, subproblem, node, value):
        if self.is_disabled(subproblem):
            return
        # print "FloatBC::expand."
        if self.active and not self.dupnodes.has_key(node):
            value += self.profileOffset
            comp = self.field.getIndex(self.field_component).integer()
            if not self.intersected:
                if self.expandfirst==None:
                    self.expandfirst = 1
                else:
                    try:
                        prev = self.field.value(node, comp)
                        self.field.setvalue(node, comp,
                                            value-self.profileStart+prev)
                    except ooferror2.ErrNoSuchField:
                        pass
            else: # Intersection case -- profileStart assumed received.
                try:
                    prev = self.field.value(node, comp)
                    self.field.setvalue(node, comp,
                                        value-self.profileStart+prev)
                except ooferror2.ErrNoSuchField:
                    pass
                

    # Build the equivalent DirichletBC for your field and equation,
    # and deactivate yourself.
    def fix(self, offset):
        self.active = 0
        return DirichletBC(
            self.field, self.field_component,
            self.equation, self.eqn_component,
            self.profile, self.boundary, offset=offset,
            user_enable=not self.explicit_disable,
            visible=False, promoted=True)
        
    #
    def intersect(self, other): # Other should be another FloatBC.
        return (self.field == other.field and
                self.field_component == other.field_component and
                self.equation == other.equation and
                self.eqn_component == other.eqn_component)

    def display(self):
        return "FloatBC / %s[%s] / %s[%s]" % (`self.field`,
                                              self.field_component,
                                              `self.equation`,
                                              self.eqn_component)

registeredclass.Registration(
    "Floating",
    BC,
    FloatBC,
    ordering=3,
    params= bc_field_param_group + bc_equation_param_group + [
        parameter.RegisteredParameter(
            'profile', reg=profile.Profile, value=None,
            tip='Profile describing how this condition varies.'),
        bc_boundary_param
        ],
    tip="Fix the values of Fields on the boundary, up to an unspecified offset.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/floating.xml')
    )


# PeriodicBC's are special in that they don't contain a profile object.
# TODO: There might be interesting ways of using profiles with
# PeriodicBC's.  For instance, a user might want to specify that
# boundaries are periodic up to an offset.

class PeriodicBC(BC):
    def __init__(self, field, field_component, equation, eqn_component,
                 boundary, user_enable=True):
        self.field = field
        self.field_component = field_component
        self.equation = equation
        self.eqn_component = eqn_component

        self.boundary = boundary
        # periodic boundary conditions must keep track of exactly two
        # boundaries the string returned by the widget will be two
        # names separated by a space
        self.boundaries = string.splitfields(self.boundary)

        self.boundary_obj = None
        self.name = None
        self.mesh_obj = None

        self.computableDict = {}
        self.explicit_disable = not user_enable
        self.visible = True
        self.subordinate = False

        # needed for meshIO
        self.profile = None
    

    def copy(self, boundary):
        return PeriodicBC(self.field, self.field_component,
                          self.equation, self.eqn_component,
                          boundary,
                          user_enable=not self.explicit_disable)

    def toggle_explicit_disable(self):
        BC.toggle_explicit_disable(self)
        for bc in self.floatBCs:
            bc.toggle_explicit_disable()

    
    # we must override a few functions from the class BC because
    # for periodic BC, the name passed in is for two boundaries
    def check(self, mesh, exclude=None):
        try:
            mesh_obj = ooflib.engine.mesh.meshes[mesh]
        except:
            return "There is no mesh named '%s'!" % mesh
        if len(self.boundaries) != 2:
            return "PeriodicBC needs exactly 2 boundaries"
        try:
            boundary_obj0 = mesh_obj.getBoundary(self.boundaries[0])
        except:
            return "There is no boundary named '%s'!" % self.boundaries[0]
        try:
            boundary_obj1 = mesh_obj.getBoundary(self.boundaries[1])
        except:
            return "There is no boundary named '%s'!" % self.boundaries[1]
        # make sure the pair of boundaries span opposite, periodic
        # outer edges with opposite directions.
        if not (boundary_obj0.whichPeriodicEdge(mesh_obj)
                and boundary_obj1.whichPeriodicEdge(mesh_obj)):
            return "Boundaries must be periodic outer edges"
        coord0, side0, dir0 = boundary_obj0.whichPeriodicEdge(mesh_obj)
        coord1, side1, dir1 = boundary_obj1.whichPeriodicEdge(mesh_obj)
        if not (coord0==coord1 and side0 != side1):
            return "Boundaries must be periodic outer edges"
        if dir0 != -dir1:
            return "Boundaries must have opposite directions"
        return (boundary_obj0.checkCondition(self, exclude=exclude) and \
                boundary_obj1.checkCondition(self, exclude=exclude))


    def add_to_mesh(self, name, mesh):
        self.name = name
        self.mesh = mesh # Mesh name.
        
        self.mesh_obj = ooflib.engine.mesh.meshes[mesh] # Who object.

        try:
            self.boundary_obj0 = self.mesh_obj.getBoundary(self.boundaries[0])
        except KeyError:
            raise
        try:
            self.boundary_obj1 = self.mesh_obj.getBoundary(self.boundaries[1])
        except KeyError:
            raise

        self.boundary_obj0.addCondition(self)
        self.boundary_obj1.addCondition(self)

        # Point boundaries for each pair of mesh nodes along the
        # boundary are stored in the mesh.  If this is the first time
        # getPeriodicPointBoundaries is called, it creates them.
        # Otherwise it just returns the stored list of point
        # boundaries.
        periodicPointBoundaries = \
            self.mesh_obj.getPeriodicPointBoundaries(self.boundary)
        
        self.floatBCs = []
        for bdy in periodicPointBoundaries:

            # Create floatBC with constant profile
            newprofile = profile.ConstantProfile(0)
            newbc = FloatBC(self.field, self.field_component, self.equation,
                            self.eqn_component, newprofile, bdy.name(),
                            visible=False, subordinate=True)
            # need unique name
            ## The string "aux_pointbdy" also appears in
            ## IO/boundarycondition.py.  See the TODOs in that file.
            aux_bc_name = self.mesh_obj.uniqueBCName("aux_pointbdy")
            newbc.add_to_mesh(aux_bc_name, mesh)
            self.floatBCs.append(newbc)

        # Auto-enable must precede the mesh_obj add call below, which
        # builds the display string, which depends on side-effects of
        # auto_enable. (Really?)
        for subproblem in self.mesh_obj.subproblems():
            self.auto_enable(subproblem)
        self.mesh_obj.addBdyCondition(name, self)


    def disconnect(self):
        self.boundary_obj0.removeCondition(self)
        self.boundary_obj0 = None
        self.boundary_obj1.removeCondition(self)
        self.boundary_obj1 = None
        self.mesh_obj = None
        self.mesh = None
        self.computableDict = {}

        for bc in self.floatBCs:
            bc.remove_from_mesh()

            
    def intersect(self, other): # Other should be another PeriodicBC.
        return (self.field == other.field and
                self.field_component == other.field_component and
                self.equation == other.equation and
                self.eqn_component == other.eqn_component)


    # Used by the GUI.
    def display_string(self):
        if self.boundary_obj0 and self.boundary_obj1:
            resstr = (self.boundary_obj0.name( ) + " "
                      + self.boundary_obj1.name( ) 
                      + " / " + self.display())
        else:
            resstr = "Unassigned / " + self.display()

        # If there's no Mesh, or if the BC isn't computable on any of
        # the Mesh's subproblems, then mark the BC as "disabled".
        if self.mesh_obj:
            for s in self.mesh_obj.subproblems():
                if not self.is_disabled(s.getObject()): 
                    return resstr
                elif self.explicit_disable:
                    disstr = "(user disabled) "
                else:
                    disstr = "(incomputable) " #better description?
        return disstr + resstr

    
    def addToBoundary(self, boundary):
        boundary.addPeriodicCondition(self)

    def checkBoundary(self, boundary, exclude):
        return boundary.checkPeriodicCondition(self, exclude)
        
    def computable(self, subproblem):
        if not self.mesh_obj:
            return None
        return subproblem.is_defined_field(self.field) and \
               subproblem.is_active_field(self.field) and \
               subproblem.is_active_equation(self.equation)

    def display(self):
        return "PeriodicBC / %s[%s] / %s[%s]" % (`self.field`,
                                              self.field_component,
                                              `self.equation`,
                                              self.eqn_component)
      
registeredclass.Registration(
    "Periodic",
    BC,
    PeriodicBC,
    ordering=5,
    params= bc_field_param_group + bc_equation_param_group + [
        bc_periodic_edge_bdy_param
        ],
    tip="Set the values of a component of a field to be equal along two opposite boundaries.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/periodic.xml')
    )

########################################################################
# # # # # # # # # # NEUMANN BOUNDARY CONDITION # # # # # # # # # # # # #
########################################################################

from ooflib.engine.profile import Location
from ooflib.SWIG.common import coord

# Wrapper function for flux boundary profiles.  Takes a number of
# floating-point values provided from within C++, translates them into
# a Location object, and calls the profile object with the location.
# OPT: If this is too slow, the Location construction process can be
# moved into the interior of the flux class's BCCallback routines,
# defined in boundarycond.swg.
def flux_locator(profile_obj, posx, posy, time, normx, normy,
                 distance, fraction):
    pos = coord.Coord(posx, posy)
    norm = coord.Coord(normx, normy)
    loc = Location(pos, normal=norm, s=distance, alpha=fraction, time=time)
    return profile_obj(loc)

#
# Actual flux boundary condition.  Simpler than the others, because
# most of the action happens in C++.
class NeumannBC(BC):
    def __init__(self, flux, profile, boundary, normal = None,
                 user_enable=True):
        self.flux = flux
        self.normal = normal  
        BC.__init__(self, profile, boundary, user_enable=user_enable)

    def copy(self, boundary):
        return NeumannBC(self.flux, self.profile, boundary,
                         normal=self.normal,
                         user_enable=not self.explicit_disable)
    
    def write(self, file, indent):
        file.write(indent+"NeumannBC("+problem.inversedict[self.flux]+")")
    def addToBoundary(self, boundary):
        boundary.addFluxCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkFluxCondition(self, exclude)

    def computable(self, subproblem):
        if not self.mesh_obj:
            return None
        return subproblem.is_active_flux(self.flux)
    
    #
    # Invoke means to do the integral over the appropriate
    # boundary, and make the contributions to the RHS.
    def __call__(self, subproblem, edgeset, profile, time):
        if self.is_disabled(subproblem):
            return
        if subproblem.is_active_flux(self.flux):
            applicator = boundarycond.NeumannBCApp(subproblem, self.flux,
                                                   edgeset)
            # For this call, "profile" must be a Python callable.
            applicator.integrate(flux_locator, profile, self.normal, time)
        else:
            raise ErrSetupError('Attempt to invoke NeumannBC on inactive flux.')
    def display(self):
        return "NeumannBC/%s" % `self.flux`
##    def display(self):
##        return "NeumannBC(flux=%s)" % `self.flux`

# NeumannBC doesn't make sense to apply point-wise, use the ForceBC for
# that.  ForceBC's are specified in terms of equations and components,
# rather than fluxes, because it makes more sense. 


registeredclass.Registration(
    "Neumann",
    BC,
    NeumannBC,
    ordering=4,
    params = [
        meshparameters.FluxParameter(
            "flux",
            tip="Flux whose value is specified across this boundary."),
        profile.FluxProfileSetParameter(
            "profile",
            tip="Profiles describing the prescribed flux."),
        bc_edge_bdy_param,
        parameter.BooleanParameter("normal",
                                   tip="If true, the boundary condition is given in a local right-handed coordinate system where x is the outward normal to the boundary and y is the tangent direction.  If false, x and y are lab coordinates."
                                   )],
    tip="Set the normal component of a Flux (eg, stress, heat flux) along the boundary.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/neumann.xml'))

########################################################################
#Interface branch

#No addToBoundary method.
#No addJumpCondition or checkJumpCondition methods in EdgeBoundary class.
class JumpBC(BC):
    def __init__(self, field, field_component,
                 jump_value, independent,
                 boundary, #This may be an interfacename
                 user_enable=True):
        self.field = field
        self.field_component = field_component
        self.jump_value=jump_value
        self.independent=independent

        BC.__init__(self,
                    profile.LinearProfile(0,jump_value),
                    boundary, user_enable=user_enable)

    def copy(self, boundary):
        return JumpBC(self.field, self.field_component,
                      self.jump_value, self.independent,
                      boundary,
                      user_enable=not self.explicit_disable)
##    def addToBoundary(self, boundary):
##        boundary.addForceCondition(self)
##    def checkBoundary(self, boundary, exclude):
##        return boundary.checkForceCondition(self, exclude)

    def computable(self, subproblem):
        if not self.mesh_obj:
            return None
        return subproblem.is_active_field(self.field)
    #
    # Applied via "call" method, just creates an applicator if it
    # doesn't already have one, and calls it with the node and
    # equation component, and the value to set.  
    #
    def __call__(self, subproblem, node, value):
        pass
##        if self.is_disabled(subproblem):
##            return
##        eqnindex = self.equation.getIndex(self.eqn_component).integer()
##        try:
##            boundarycond.applyForceBC(subproblem, self.equation,
##                                      node, eqnindex, value)
##        except ooferror2.ErrNoSuchField: # eqn not active at that node
##            pass
    def display(self):
        return ("JumpBC / %s[%s] / jumpvalue=%s / independent=%s" 
                % (`self.field`,
                   self.field_component,
                   self.jump_value,
                   self.independent))
                
    #Override the one from the parent BC class
    def add_to_mesh(self, name, mesh):
        self.name = name 
        self.mesh = mesh # Mesh name.
        
        self.mesh_obj = ooflib.engine.mesh.meshes[mesh] # Who object.

        femeshobj=self.mesh_obj.getObject()
        femeshobj.addInterfaceBC(name)

##        try:
##            self.boundary_obj = self.mesh_obj.getBoundary(self.boundary)
##        except KeyError:
##            raise

        self.profile_obj.addCondition(self)
##        self.boundary_obj.addCondition(self)

        # Auto-enable must precede the mesh_obj add call below, which
        # builds the display string, which depends on side-effects of
        # auto_enable. (Really?)
        for subproblem in self.mesh_obj.subproblems():
            self.auto_enable(subproblem)
        self.mesh_obj.addBdyCondition(name, self)

    #Override the one from the parent BC class
    def remove_from_mesh(self):
        self.mesh_obj.rmBdyCondition(self)

    #Override the one from the parent BC class
    def disconnect(self):
        self.profile_obj.removeCondition(self)
        self.mesh_obj = None
        self.mesh = None
        self.computableDict = {}
        #self.boundary_obj.removeCondition(self)
        #self.boundary_obj = None

        # Break circular references, but leave in the ones
        # corresponding to parameters.
        
        # self.boundary = None
        # self.profile=None
        self.profile_obj = None
        # Clean up callbacks.
        map(switchboard.removeCallback, self.sw_cb)

    def intersect(self, other):
        # Other should be another JumpBC.
        return (self.field == other.field and
                self.field_component == other.field_component)

    #Override the one from the parent BC class
    def check(self, mesh, exclude=None):
        # Check that this bc doesn't conflict with any other bc on
        # this mesh. This doesn't check the *values* of the boundary
        # conditions.  It only checks that two Dirichlet or floating
        # boundary conditions don't apply to the same degrees of
        # freedom on the same boundary.
        # If the 'exclude' variable is equal to some other bc, then
        # conflicts with that bc aren't counted.

        try:
            meshctxt = ooflib.engine.mesh.meshes[mesh]                
        except:
            return "There is no mesh named '%s'!" % mesh

        meshobj=meshctxt.getObject()
        for jumpbcname in meshobj.listInterfaceBC():
            jumpbc=meshctxt.getBdyCondition(jumpbcname)
            if jumpbc.boundary==self.boundary and \
                   jumpbc is not exclude and \
                   jumpbc.intersect(self) and \
                   not jumpbc.is_explicitly_disabled():
                return "Jump BC collision on interface"
##        try:
##            boundary_obj = mesh_obj.getBoundary(self.boundary)
##        except:
##            return "There is no boundary named '%s'!" % self.boundary
##        return boundary_obj.checkCondition(self, exclude=exclude)
        return ""

registeredclass.Registration("Field Discontinuity",
                             BC,
                             JumpBC,
                             ordering=6,
                             params = bc_field_param_group + [
##    parameter.RegisteredParameter(
##    'profile', reg=profile.Profile, value=None,
##    tip='Profile describing how this condition varies.'),
    parameter.FloatParameter('jump_value', 0.0,
                             tip='Value for the field jump or discontinuity'),
    parameter.BooleanParameter('independent', 0,
                               tip='Ignore jump value?'),
##    meshparameters.MeshEdgeBdyInterfaceParameter('boundary',
##                                                 tip='Edge boundary or interface to which this condition applies.')
    bc_edge_bdy_param],
                             tip="Jump in a field value across a boundary",
                             discussion = 
"""<para>

This boundary condition imposes a finite jump in the value of a field
across an interface or boundary. The 'sense' of the jump refers to the
field value on the node to the right of the boundary relative to the
field value on the node to the left of the boundary. If independent is
set to True, then no constraints are specified between the field
values.  When no boundary conditions for the field are specified
across an interface, hidden Floating boundary conditions are
automatically created that keep the field value continuous across the
boundary when the finite element system is solved.

</para>"""
                             )
