// -*- C++ -*-
// $RCSfile: csubproblem.C,v $
// $Revision: 1.26 $
// $Author: langer $
// $Date: 2008-06-10 21:05:04 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <map>
#include <utility>		// for std::pair
#include <vector>

#include "common/lock.h"
#include "common/IO/progressbar.h"
#include "common/printvec.h"	// debugging
#include "common/tostring.h"
#include "engine/cconjugate.h"
#include "engine/csrmatrix.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/linearizedsystem.h"
#include "engine/meshiterator.h"
#include "engine/nodalequation.h"
//AMR subproblem
#include "engine/cscpatch.h"
#include "engine/nodalfluxes.h"
#include "engine/nodalscpatches.h"
#include "engine/node.h"
#include "engine/csubproblem.h"

typedef std::pair<FuncNode::FieldSet, FuncNode::EquationSet> FEPair;
typedef std::map<FEPair, std::vector<int> > LocalMapDict;

long CSubProblem::globalCSubProblemCount = 0;
SLock globalCSubProblemCountLock;

CSubProblem::CSubProblem()
  : rwlock(0),
    mesh(0),
    precomputeRequired(false),
    linearsystem(this),
    global_equation_map(0),
    n_active_eqn(0),
    n_active_field(0),
    local_time(0.0)
{
  globalCSubProblemCountLock.acquire();
  ++globalCSubProblemCount;
  globalCSubProblemCountLock.release();
}

CSubProblem::~CSubProblem() {
  //AMR subproblem
  std::map<const int, NodalSCPatches*>::iterator piter;
  for(piter=scpatches.begin(); piter!=scpatches.end(); piter++)
    delete piter->second;

  std::map<const int, NodalFluxes*>::iterator fiter;
  for(fiter=recovered_fluxes.begin(); fiter!=recovered_fluxes.end(); fiter++)
    delete fiter->second;

  globalCSubProblemCountLock.acquire();
  --globalCSubProblemCount;
  globalCSubProblemCountLock.release();
}

long get_globalCSubProblemCount() {
  return CSubProblem::globalCSubProblemCount;
}

CMicrostructure *CSubProblem::get_microstructure() const {
  return mesh->get_microstructure();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// NodeIterator CSubProblem::node_iterator() const {
//   return NodeIterator(new SubProblemNodeIterator(this));
// }

// FuncNodeIterator CSubProblem::funcnode_iterator() const {
//   return FuncNodeIterator(new SubProblemFuncNodeIterator(this)); 
// }

// ElementIterator CSubProblem::element_iterator() const {
//   return ElementIterator(new SubProblemElementIterator(this));
// }

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CSubProblem::EquationData::EquationData()
  : active(0),
    kin_active(0),
    dyn_active(0)
{}

void CSubProblem::activate_equation(Equation &eqn) {
  // Trace("CSubProblem::activate_equation " + eqn.name());
  int index = eqn.index();
  // Make sure that there's an EquationData entry for this Equation
  if(index >= int(eqndata.size())) {
    eqndata.resize(index+1);
  }
  if(!eqndata[index].active) {
    // Mark the Equation as active
    eqndata[index].active = true;
    // Offset of this Equation in the Nodes' NodalEquation lists
//     eqndata[index].active_index = n_active_eqn;
    // Total number of active NodalEquations at each Node.
    ++n_active_eqn;

    // Create NodalEquations for this Equation in each Node.  This
    // also updates the Nodes' EquationSets.
    for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
      ni.node()->addEquation(mesh, eqn);
    }
    // Activate fluxes required for this Equation
    eqn.activate_fluxes(this);
  }
}

void CSubProblem::deactivate_equation(Equation &eqn) {
  if(is_active_equation(eqn)) {
    int index = eqn.index();
    eqndata[index].active = false;
    for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
      ni.node()->removeEquation(mesh, eqn);
    }
    // perform garbage collection on the nodaleqn list
    mesh->clean_nodaleqn();
    --n_active_eqn;
     eqn.deactivate_fluxes(this);
  }
}

// TODO PLASTICITY: How is is_active_equation used?  Should it examine
// each Node's EquationSet, or just check the CSubProblem's
// EquationData.active?  This may need a closer look when we get
// automatically activated auxilliary equations, esp. in
// Flux::make_stiffness and related functions.  Most other invocations
// of is_active_equation are used to determine what equations the user
// has explicitly activated.

bool CSubProblem::is_active_equation(const Equation &eqn) const {
  std::vector<EquationData>::size_type index = eqn.index();
  return (index < eqndata.size()) && eqndata[index].active;
}

int CSubProblem::n_active_eqns() const {
  return n_active_eqn;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Kinetics stuff

void CSubProblem::kinetic_activate_equation(Equation &eqn) {
  if (eqn.get_kinetic() >0)
    eqndata[eqn.index()].kin_active = true;
  //    eqn.activate_kinetics();
}


void CSubProblem::kinetic_deactivate_equation(Equation &eqn) {
  if(is_kinetically_active_equation(eqn) && eqn.get_kinetic() >0)
    //eqn.deactivate_kinetics();
    eqndata[eqn.index()].kin_active = true;
}


bool CSubProblem::is_kinetically_active_equation(Equation &eqn) {
  return is_active_equation(eqn) && eqndata[eqn.index()].kin_active;
}

bool CSubProblem::kinetics_activated() {
  std::vector<Equation*> *eqlist = all_equations();
  for(std::vector<Equation*>::size_type i=0; i< eqlist->size(); i++) {
    Equation * eqn = (*eqlist)[i];
    if (is_kinetically_active_equation(*eqn))
      {
	delete eqlist;
	return true;
      }
  }
  delete eqlist;
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Dynamics stuff

void CSubProblem::activate_dynamics(Equation &eqn) {
  if (eqn.get_dynamics() >0)
    //eqn.activate_dynamics();
    eqndata[eqn.index()].dyn_active = true;
}


void CSubProblem::deactivate_dynamics(Equation &eqn) {
  if(is_dynamically_active_equation(eqn) && eqn.get_dynamics() >0)
    //eqn.deactivate_dynamics();
    eqndata[eqn.index()].dyn_active = false;
}


bool CSubProblem::is_dynamically_active_equation(Equation &eqn) {
  return is_active_equation(eqn) && eqndata[eqn.index()].dyn_active;
}

bool CSubProblem::dynamics_activated() {
  std::vector<Equation*> *eqlist = all_equations();
  for(std::vector<Equation*>::size_type i=0; i<eqlist->size(); i++) {
    Equation * eqn  = (*eqlist)[i];
    if (is_dynamically_active_equation(*eqn))
      {
	delete eqlist;
	return true;
      }
  }
  delete eqlist;
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Flux bookkeeping. 

void CSubProblem::activate_flux(const Flux &flux) {
  std::vector<int>::size_type index = flux.index();
  if(index >= active_flux.size()) {
    active_flux.resize(index+1, 0);
  }
  ++active_flux[index];
}

void CSubProblem::deactivate_flux(const Flux &flux) {
  std::vector<int>::size_type index = flux.index();
  if(index < active_flux.size()) {
    if(active_flux[index] > 0)
      --active_flux[index];
  }
}

bool CSubProblem::is_active_flux(const Flux &flux) const {
  std::vector<int>::size_type index = flux.index();
  if(index < active_flux.size()) {
    return active_flux[index] > 0;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Field stuff

// activate, etc, for Fields have to be routed through Field virtual
// functions, because they have to know if the Field is a
// CompoundField or not.  The routines that do the real work are
// FEMesh::do_activate, etc.

CSubProblem::FieldData::FieldData()
  : active(0),
    defined(0)
{}

Field &CSubProblem::define_field(Field &field) {
  // Trace("CSubProblem::define_field " + field.name());
  field.define(this);
  return field;
}

void CSubProblem::undefine_field(Field &field) {
//   Trace("CSubProblem::undefine_field " + field.name());
  field.undefine(this);
}

void CSubProblem::activate_field(Field &field) {
  // Trace("CSubProblem::activate_field " + field.name());
  field.activate(this);
}

void CSubProblem::deactivate_field(Field &field) {
//   Trace("CSubProblem::deactivate_field " + field.name());
  field.deactivate(this);
}

void CSubProblem::do_define_field(Field &field) {
  // Trace("CSubProblem::do_define_field " + field.name() + " " + to_string(field.index()));
  requirePrecompute();
  std::vector<FieldData>::size_type index = field.index();
  if(fielddata.size() <= index) {
    fielddata.resize(index+1);
  }
  FieldData &fdata = fielddata[index];
  if(!fdata.defined) {
    fdata.defined = true;
    for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
      ni.node()->addField(mesh, field);
    }
  }
}

void CSubProblem::do_undefine_field(Field &field) {
//   Trace("CSubProblem::do_undefine_field " + field.name())
  requirePrecompute();
  if(is_defined_field(field)) {
    if(is_active_field(field)) {
      deactivate_field(field);
    }
    std::vector<FieldData>::size_type index = field.index();
    if(index < fielddata.size()) {
      fielddata[index].defined = false;
      for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
	ni.node()->removeField(mesh, field);
      }
      mesh->clean_doflist();
    }
  }
}

// Function to copy a field from another CSubProblem.  The other
// CSubProblem must have the same geometry, i.e. the same-indexed
// funcnodes must be in the same location, and the field must be
// defined.  Called by meshmenu.copyMesh.

void CSubProblem::acquire_field_data(Field &field, const CSubProblem *other) {
  for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i) {
    for(int d=0; d<field.ndof(); ++d) {
      if(i.node()->hasField(field)) {
	field(i.node(), d)->value() = 
	  field(other->mesh->getFuncNode(i.count()), d)->value();
      }
    }
  }
}

bool CSubProblem::is_defined_field(const Field &field) const {
  return (field.index() < fielddata.size())
    && fielddata[field.index()].defined;
}

void CSubProblem::do_activate_field(Field &field) {
  if(!is_active_field(field)) {
    n_active_field++;
    FieldData &fdata = fielddata[field.index()];
    fdata.active = true;
    // The flag in the DegreeOfFreedom objects is *not* set here.
    // It's set only during the solution process (by
    // CSubProblem::mapdofeqs()).  That's because DegreeOfFreedoms are
    // shared among subproblems, but activity is a per-subproblem
    // trait.
  }
}

void CSubProblem::do_deactivate_field(Field &field) {
  if(is_active_field(field)) {
    n_active_field--;
    fielddata[field.index()].active = false;
  }
}

bool CSubProblem::is_active_field(const Field &field) const {
  return (field.index() < fielddata.size()) && fielddata[field.index()].active;
}

int CSubProblem::n_active_fields() const {
  return n_active_field;
}

// void CSubProblem::set_in_plane(const Field &field, bool planarity) {
//   if(field.index() >= fielddata.size()) {
//     fielddata.resize(field.index()+1);
//   }
//   fielddata[field.index()].in_plane = planarity;
// }

bool CSubProblem::in_plane(const Field &field) const {
  return mesh->in_plane(field);
//   return (field.index() < fielddata.size())
//     && fielddata[field.index()].in_plane;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Return a list of all the defined compound fields.  Compound fields
// are the ones that correspond directly to physical fields.

std::vector<CompoundField*>* CSubProblem::all_compound_fields() const {
  std::vector<CompoundField*>* flist = new std::vector<CompoundField*>;
  std::vector<CompoundField*> &allfields = CompoundField::allcompoundfields();
  flist->reserve(fielddata.size());
  for(std::vector<CompoundField*>::size_type i=0; i<allfields.size(); i++) {
    // TODO PLASTICITY: This only picks up explicitly defined Fields.
    // Should it look at Node::fieldset instead?  Should there be
    // another function that retrieves implicit Fields that may not be
    // in CSubProblem::fielddata?
    if(is_defined_field(*allfields[i])) 
      flist->push_back(allfields[i]);
  }
  return flist;
}

// Likewise for Equation objects.
std::vector<Equation*>* CSubProblem::all_equations() const {
  std::vector<Equation*> *eqlist = new std::vector<Equation*>;
  std::vector<Equation*> &alleqns = Equation::all();
  eqlist->reserve(eqndata.size());
  for(std::vector<Equation*>::size_type i=0; i<alleqns.size(); i++) {
    if(is_active_equation(*alleqns[i]))
      eqlist->push_back(alleqns[i]);
  }
  return eqlist;
}

std::vector<Flux*>* CSubProblem::all_fluxes() const {
  std::vector<Flux*> *fluxlist = new std::vector<Flux*>;
  std::vector<Flux*> &allfluxen = Flux::allfluxes();
  for(std::vector<Flux*>::size_type i=0; i<allfluxen.size(); i++) {
    if(is_active_flux(*allfluxen[i]))
      fluxlist->push_back(allfluxen[i]);
  }
  return fluxlist;
}

std::vector<Flux*> CSubProblem::allFluxes() const {  // non-swigged version.
  std::vector<Flux*> fluxlist;
  std::vector<Flux*> &allfluxen = Flux::allfluxes();
  for(std::vector<Flux*>::size_type i=0; i<allfluxen.size(); i++) {
    if(is_active_flux(*allfluxen[i]))
      fluxlist.push_back(allfluxen[i]);
  }
  return fluxlist;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void CSubProblem::make_linear_system() {
//   Trace("CSubProblem::make_linear_system");
  Progress * localbar = getProgress();

  //This call to housekeeping may be redundant
  //(see set_trivial_equation_mapping, set_equation_mapping and
  //deactivate_equation).
  mesh->housekeeping();
  // std::cerr << "nodal equations =" << nodaleqn.size() << std::endl;
  // std::cerr << "dofs =" << dof.size() << std::endl;

  // Initialize the full linear system -- sets up all three matrices.
  // TODO DEVEL: Mindlessly clearing it here makes it work like the
  // old code.  In general, one wants to keep track of whether or not
  // one is evaluating at a new time, and only clear when you're sure
  // the time-step algorithm won't need the old data.  Or maybe the
  // time-stepper itself could keep the old data.
  linearsystem.clear();
  linearsystem.initialize(mesh->nodaleqn.size(), mesh->dof.size());
  if (kinetics_activated()) 
    linearsystem.set_kinetic();
  if (dynamics_activated())
    linearsystem.set_dynamic();

  for(ElementIterator ei=element_iterator(); !ei.end(); ++ei) {
    ei.element()->make_linear_system(this, linearsystem);
    localbar->set_progress(float(ei.count()+1)/float(ei.size()));
    localbar->set_message("creating linear system: " +
			  to_string(ei.count()+1) + "/" + to_string(ei.size()));
  }
//  std::cerr << "CSubProblem::make_linear_system:\n" << linearsystem.get_stiffness() << std::endl;

  //Interface branch
  //TODO: Write an EdgementIterator for the subproblem.
  for(std::vector<Edgement*>::size_type i=0;i<mesh->edgement.size();i++)
    {
      if(mesh->edgement[i]->isSubProblemEdgement(this))
	{
	  mesh->edgement[i]->make_linear_system(this, linearsystem);
	}
    }

}

// Preliminary -- no housekeeping, just run the element's
// post-equilibration loop.
void CSubProblem::post_process() {
  //Thread control starts here
  Progress * localbar = getProgress(); //not a law firm
  //Thread control ends here
  for(ElementIterator ei=element_iterator(); !ei.end(); ++ei) {
    ei.element()->post_process(this);
    localbar->set_progress(float(ei.count()+1)/float(ei.size()));
    localbar->set_message("postprocessing: " + to_string(ei.count()+1) + "/"
			  + to_string(ei.size()));
  }

  //Interface branch
  //TODO: Write an EdgementIterator for the subproblem.
  for(std::vector<Edgement*>::size_type i=0;i<mesh->edgement.size();i++)
    {
      if(mesh->edgement[i]->isSubProblemEdgement(this))
	{
	  mesh->edgement[i]->post_process(this);
	}
    }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Routines used to pick out rows and columns of the stiffness matrix
// when extracting submatrices.

bool CSubProblem::solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  DegreeOfFreedom *dof = mesh->dof[i];
  return linearsystem.is_nonempty_col(i) && dof->isfree() && dof->active();
}

bool CSubProblem::dont_solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  DegreeOfFreedom *dof = mesh->dof[i];
  return linearsystem.is_nonempty_col(i) && (!dof->isfree() || !dof->active());
}

bool CSubProblem::solve_this_nodal_eqn(int j) const {
  //nodaleqn corresponds to unsymmetric system
  NodalEquation *ndq = mesh->nodaleqn[j];
  bool result = linearsystem.is_nonempty_row(j) && !ndq->is_dependent();
  return result;
}

bool CSubProblem::dont_solve_this_nodal_eqn(int j) const {
  return linearsystem.is_nonempty_row(j) && mesh->nodaleqn[j]->is_dependent();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix
// The conceptual subdivision is this:
//
//                 free      fixed
//                          |
//   independent      A     | C
//                          |
//                ----------+---     --- 
//     dependent      C'    | B
// 

// Rowmapper and Colmapper.
// Build a map of the rows i of the stiffness matrix for
// which f(i) is true, and return it.  Call with f(i)
// discriminating between independent (known-force) vs.
// dependent (unknown-force) equations for rows, or
// with f(i) discriminating between fixed and free
// DOFs for columns.
//
// Do this by manipulating the pre-existing mapping list.
void CSubProblem::rcmapper(bool (CSubProblem::*f)(int) const,
			   std::vector<int> &rcmap, int len) const
{
  //  Trace("FEMesh::rcmapper");
  int nrc = 0;			// # of rows or columns in the mapped submatrix
  rcmap.resize(len); 
  for(int i=0; i<len; i++) {
    if ((this->*f)(i)) {
      rcmap[i] = nrc++;
    }
    else {
      rcmap[i] = -1; // -1 is default for rows/columns which don't occur.
    }
  }
}

// Run the maps, creating the lists used to partition the stiffness
// matrix.

void CSubProblem::mapdofeqs() {
//   Trace("CSubProblem::mapdofeqs");

  // Set the "active" flag in the dofs.  This flag is used by
  // solve_for_this_dof and dont_solve_for_this_dof.  Setting the flag
  // here is a hack which will fail if more than one subproblem is
  // solved simultaneously.  But solving more than one subproblem
  // simultaneously is problematic anyway, if they share dofs.
  std::vector<Field*> &fields = Field::all();
  for(unsigned int i=0; i<fields.size(); i++) {
    Field &field = *fields[i];
    if(is_active_field(field)) {
      for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
	if(ni.node()->hasField(field)) {
	  for(int i=0; i<field.ndof(); i++) {
	    field(ni.node(), i)->activate();
	  }
	} // if node has field
      }	// loop over nodes
    } // if field is active
    else {			// field is not active
      for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
	if(ni.node()->hasField(field)) {
	  for(int i=0; i<field.ndof(); i++) {
	    field(ni.node(), i)->deactivate();
	  }
	} // if node has field
      }	// loop over nodes
    } // if field is not active
  } // loop over fields

  rcmapper(&CSubProblem::solve_this_nodal_eqn, indepeqnmap, 
	   linearsystem.nrows());
  // std::cerr << "  indepeqnmap= " << indepeqnmap << std::endl;
  rcmapper(&CSubProblem::dont_solve_this_nodal_eqn, depeqnmap, 
	   linearsystem.nrows());
  // std::cerr << "    depeqnmap= " << depeqnmap << std::endl;
  rcmapper(&CSubProblem::solve_for_this_dof, freedofmap, 
	   linearsystem.ncols()); 
  // std::cerr << "   freedofmap= " << freedofmap << std::endl;
  rcmapper(&CSubProblem::dont_solve_for_this_dof, fixeddofmap, 
	   linearsystem.ncols()); 
  // std::cerr << "  fixeddofmap= " << fixeddofmap << std::endl;
} 

// Undo BC invocation by resetting dofs and nodalequations to be free
// and independent.  Does not modify "active" flags.
void CSubProblem::resetdofeqs() {
  for(std::vector<DegreeOfFreedom*>::size_type i=0;i<mesh->dof.size();i++) 
    if(linearsystem.is_nonempty_col((int) i))
      mesh->dof[i]->free();
  for(std::vector<NodalEquation*>::size_type i=0;i<mesh->nodaleqn.size();i++)
    if(linearsystem.is_nonempty_row((int) i))
      mesh->nodaleqn[i]->independent();
}

// The cleanlists operation restores the row and column maps to 
// consistency after the application of a FloatBC.  The FloatBC's
// cause several rows or columns of the original stiffness matrix
// to be mapped (summed, actually) to a single row or column of
// the extracted 'A' matrix used in the actual linear-algebra solver.
// As a result, the range of the maps can develop holes -- cleanlists()
// closes those holes, ensuring that the range of the map is a 
// contiguous group of integers from 0 to the appropriate N.
void CSubProblem::cleanlists() {
  cleanmap(indepeqnmap);
  cleanmap(depeqnmap);
  cleanmap(freedofmap);
  cleanmap(fixeddofmap);
}

// This is a private routine that just does one map, called
// from FEMesh::cleanlists(). 
void CSubProblem::cleanmap(std::vector<int>& rcmap) {
  std::vector<int> histogram(rcmap.size(),0);
  std::vector<int> subtractor(rcmap.size(),0);

  // Make the histogram.
  for(std::vector<int>::size_type i=0;i<rcmap.size();i++) {
    if (rcmap[i]!=-1)
      histogram[ rcmap[i] ]++;
  }
  // To actually clean the list, the technique is this:
  // Move along the histogram, counting the number of zero
  // entries you see.  When you get to a nonzero entry, 
  // make an entry in the subtraction map -- the amount
  // to subtract from the value you're looking at is the
  // current zero count.  Then, make a pass through the
  // subtraction map and do the actual subtractions.
  int zcount = 0;
  for(std::vector<int>::size_type i=0 ; i<rcmap.size(); i++) {
    if(histogram[i]==0) {
      zcount++;
    }
    else {
      subtractor[i] = zcount;
    }
  }

  // Now do the subtraction.
  for(std::vector<int>::size_type i=0; i<rcmap.size(); i++) {
    if( rcmap[i] != -1 ) {
      rcmap[i] -= subtractor[ rcmap[i] ];
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void CSubProblem::get_unknowns(VECTOR_D &x, const std::vector<int> &map) const {
  mesh->get_unknowns(x, map);
}

void CSubProblem::set_unknowns(const VECTOR_D &x, const std::vector<int> &map) {
  mesh->set_unknowns(x, map);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix.  Mapping arrays
// should be set up by the time you call this.

const SparseLinkMatrix<double> &CSubProblem::stiffness_matrix() const {
  if(!linearsystem.is_initialized()) {
    throw ErrProgrammingError("Stiffness matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return linearsystem.get_stiffness();
}

const SparseLinkMatrix<double> &CSubProblem::conductivity_matrix() const {
  if(!linearsystem.is_kinetic()) {
    throw ErrProgrammingError("Conductivity matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return linearsystem.get_kinetic();
}
const SparseLinkMatrix<double> &CSubProblem::mass_matrix() const {
  if(!linearsystem.is_dynamic()) {
    throw ErrProgrammingError("Mass matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return linearsystem.get_dynamic();
}



// TODO DEVEL: These may want to move out of CSubProblem and into
// LinearSystem.
CSRmatrix CSubProblem::Amatrix() const {
//   Trace("CSubProblem::Amatrix");
  return linearsystem.stiffness_submatrix(indepeqnmap, freedofmap);
}

CSRmatrix CSubProblem::Bmatrix() const {
//   Trace("CSubProblem::Bmatrix");
  return linearsystem.stiffness_submatrix(depeqnmap, fixeddofmap);
}

CSRmatrix CSubProblem::Cmatrix() const {
//   Trace("CSubProblem::Cmatrix");
  return linearsystem.stiffness_submatrix(indepeqnmap, fixeddofmap);
}

CSRmatrix CSubProblem::Cpmatrix() const {
//   Trace("CSubProblem::Cpmatrix");
  return linearsystem.stiffness_submatrix(depeqnmap, freedofmap);
}

CSRmatrix CSubProblem::ConductivityMatrix() const {
  Trace("FEMesh::ConductivityMatrix");
  return linearsystem.kinetic_submatrix(indepeqnmap, freedofmap);
}

CSRmatrix CSubProblem::MassMatrix() const {
  Trace("FEMesh::MassMatrix");
  return linearsystem.dynamic_submatrix(indepeqnmap, freedofmap);
}

// Return the "blocking" data -- this is a map which is -1 if the
// corresponding row/column is in-plane, and is the node number if the
// corresponding row/column is out-of-plane.  Used by the block
// preconditioner, and by the Kinetic time stepping code.

// TODO: It might be more clever if these returned pointers/references
// to objects in either the subproblem or stiffness matrix classes.
std::vector<int> CSubProblem::A_rowblocks() const {
  return linearsystem.rowgroups(indepeqnmap);
}

std::vector<int> CSubProblem::A_colblocks() const {
  return linearsystem.colgroups(freedofmap);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void CSubProblem::get_rhs(VECTOR_D &rhs, const std::vector<int> &map) const {
//   Trace("FEMesh::get_rhs");
  VECTOR_D &body_rhs = linearsystem.get_body_rhs();
  VECTOR_D &boundary_rhs = linearsystem.get_boundary_rhs();
  for(std::vector<NodalEquation*>::size_type i=0; i<mesh->nodaleqn.size(); i++)
    {
      if(map[i] != -1) {
	rhs[map[i]] += body_rhs[i] + boundary_rhs[i];
	// Merge nodal equation values with boundary RHS contribs.
      }
    }
}

VECTOR_D &CSubProblem::boundary_rhs() const {
  if(!linearsystem.is_initialized()) {
    throw ErrProgrammingError("Boundary right hand side not available!",
			      __FILE__, __LINE__);
  }
  return linearsystem.get_boundary_rhs();
}

void CSubProblem::clear_brhs() {
  linearsystem.clear_brhs();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Routines for symmetrizing the matrix (or trying to), using
// conjugacy data.

// In the case of inconsistent conjugacy data, build a trivial
// equation map.
void CSubProblem::set_trivial_equation_mapping() {
  if (global_equation_map) {
    delete global_equation_map;
  }

  //This call to housekeeping may be redundant (see deactivate_equation).
  // Make sure the lists are contiguous.
  mesh->housekeeping();

  global_equation_map = new std::vector<int>(mesh->nodaleqn.size());
  for(unsigned int i=0; i<global_equation_map->size(); ++i) {
    (*global_equation_map)[i]=i;
  }
}

// See if a local conjugacy map has already been constructed for the
// given fieldset and nodeset, and construct a new one if necessary.
// (In this context, "local" means "local to a Node".)  The local map
// is a vector of ints.  For the i^th equation in a FuncNode's list of
// NodalEquations, localmap[i] is the index of the conjugate degree of
// freedom in the Node's list of DegreeOfFreedoms.  If there is no
// conjugate degree of freedom, localmap[i] is -1.

static std::vector<int> &getLocalMap_(FuncNode::FieldSet &fieldset,
				      FuncNode::EquationSet &eqnset,
				      LocalMapDict &localmaps,
				      CSubProblem *subproblem,
				      const std::vector<CConjugatePair*> &pairs)
{
  FEPair key(fieldset, eqnset);
  LocalMapDict::iterator i = localmaps.find(key);
  if(i == localmaps.end()) {
    // create a new map
    std::vector<int> lmap(eqnset.dimsum(), -1); // rub it and a gneie appears
    for(std::vector<CConjugatePair*>::const_iterator p=pairs.begin();
	p < pairs.end(); ++p)
      {
	const Field &field = *(*p)->get_field();
	const Equation &eqn = *(*p)->get_equation();
	if(subproblem->is_active_field(field) &&
	   subproblem->is_active_equation(eqn))
	  {
	    const FieldIndex &equationcomp = *(*p)->get_equation_component();
	    const FieldIndex &fieldcomp = *(*p)->get_field_component();
	    int eindx = eqnset.offset(&eqn) + equationcomp.integer();
	    int findx = fieldset.offset(&field) + fieldcomp.integer();
	    lmap[eindx] = findx;
	  }
      }
    localmaps.insert(LocalMapDict::value_type(key, lmap)); // copies lmap
    return localmaps[key];	// returns reference to copy in localmaps
  }
  else {
    return (*i).second;		// reuse an existing map
  }
}


void CSubProblem::set_equation_mapping(
			  const std::vector<CConjugatePair*> *conjugacylist)
{
  // Build a mapping array such that, when you traverse
  // FEMesh::nodaleqn[map[i]] for i from 0 to FEMesh::nodaleqn.size(),
  // the nodal equations come up in the same order as their conjugate
  // degrees of freedom do when you index that array (FEMesh::dof)
  // directly.  Store this map in the global_equation_map variable.
  // Using this ordering at stiffness-matrix-construction time causes
  // the resulting stiffness matrix to be as symmetric as possible.

  // Any equations for which conjugate DOFs do not exist come at the
  // tail of this list in no particular order.

  // For any given i, and in a given node's lists, the cols[i]-th
  // degree of freedom in the node is conjugate to the rows[i]-th
  // nodal equation in the node, with i ranging over the number of
  // active equations or degrees of freedom -- the number of *active*
  // eqns/dofs must match, but there may be inactive dofs.  There are
  // never inactive equations.

  if (global_equation_map) {
    delete global_equation_map;
  }
  //This call to housekeeping may be redundant (see deactivate_equation).
  // Make sure the lists are contiguous.
  mesh->housekeeping();
  global_equation_map = new std::vector<int>(mesh->nodaleqn.size(), -1);

  LocalMapDict localmaps;
  int max_target_row = -1;
  for(FuncNodeIterator iter=funcnode_iterator(); !iter.end(); ++iter) {
    FuncNode *node = iter.node();

    std::vector<NodalEquation*> &eqnlist = node->eqnlist;
    // Get the localmap, which indicates the conjugacy relations
    // between nodal equations and degrees of freedom in the Node.
    std::vector<int> &localmap = getLocalMap_(node->fieldset,
					      node->equationset,
					      localmaps, this, *conjugacylist);
    for(std::vector<NodalEquation*>::size_type i=0; i<eqnlist.size(); ++i) {
      // Only do stuff if this mapping makes sense.
      if (localmap[i]!=-1) {
	NodalEquation* neq = eqnlist[i];
	int source_row = neq->ndq_index(); // Index in master stiffness mtx

	DegreeOfFreedom *dof = (node->doflist)[localmap[i]];
	int target_row = dof->dofindex();  // Also in master stiffness mtx
	
	(*global_equation_map)[source_row] = target_row;

	if(target_row > max_target_row)
	  max_target_row = target_row;
      }
    }
  }
  
  // There may be unmapped equations due to incomplete conjugacy data.
  // Also, there may be equations for which there are no degrees of
  // freedom because they're going to be dependent equations.  In
  // either case, these "extra" equations will not have been mapped
  // anywhere.  Make a second pass through the list, and assign any
  // unassigned equations an incremental index, starting with
  // max_target_row.  Bounds are not an issue, because of the
  // range-compression step which follows this block.
  for(FuncNodeIterator iter=funcnode_iterator(); !iter.end(); ++iter) {
    FuncNode *node = iter.node();
    for(std::vector<NodalEquation*>::size_type i=0; 
	i<(node->eqnlist).size(); ++i) {
      NodalEquation *neq = (node->eqnlist)[i];
      int source_index = neq->ndq_index(); 
      if ( (*global_equation_map)[source_index] == -1 ) {
	max_target_row++;
	(*global_equation_map)[source_index]=max_target_row;
      }
    }
  }

  // Now fix the range of the global_equation_map -- the master
  // stiffness matrix can be wider than it is tall, so some of the
  // equations could be out of range.  Fix this by compressing the
  // range but preserving the order.

  // First, build a map indexed by the *range* elements of the map,
  // but preserving the domain information.  This is the "inverse" of
  // the forward-map.
  std::map<int,int> back_map;
  for(std::vector<int>::size_type i=0; i<global_equation_map->size(); ++i) {
    back_map[(*global_equation_map)[i]]=i;
  }
  // Then go through the backward map in range-of-forward-map
  // (domain-of-backward-map, key of std::map structure) order, and
  // for each element of the domain of the forward map (range of the
  // backward map, value of std::map structure), assign consecutive
  // integers to the range of the forward map.
  int key_count=0;
  // Note: This loop traverses back_map sorted in the order
  // according to the keys (bmi->first).
  for(std::map<int,int>::const_iterator bmi=back_map.begin();
      bmi!=back_map.end(); ++bmi)
    {
      (*global_equation_map)[(*bmi).second]=key_count;
      key_count++;
    }

} // set_equation_mapping

void CSubProblem::clear_symmetricity_maps()
{
  //destroy the symmetricity maps after using them
  delete global_equation_map;
  global_equation_map=0;
}

// Having remapped the equation array, we need a helper function so
// that boundary conditions can set the appropriate equations to be
// dependent.  

const NodalEquation &CSubProblem::map_forward_equation(const Equation &equation,
						       const FuncNode &node,
						       const int &eqcomp) const
{
  int target_index = map_forward_index(equation, node, eqcomp);
  return *(mesh->nodaleqn[target_index]);
}

int CSubProblem::map_forward_index(const Equation & equation, 
				   const FuncNode &node, const int &eqcomp)
  const 
{
  // Reads the global equation map -- see set_equation_mapping for details.
  int original_row = equation.nodaleqn(node, eqcomp)->ndq_index();
  return (*global_equation_map)[original_row];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// dump_dof and dump_eqn are commented out because they require const
// iterators,which I'm too lazy to write just now.

// void CSubProblem::dump_dof() const {
//   for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i) {
//     const FuncNode &node = *i.node();
//     std::cout << "node = " << node << std::endl;
//     for(std::vector<Field*>::size_type f=0; f<Field::all().size(); f++) {
//       const Field &field = *Field::all()[f];
//       if(field.is_defined(this)) {
// 	for(int j=0; j<field.ndof(); j++) {
// 	  std::cout << "\t" << field.name() << "[" << j << "] dof="
// 	       << *field(node, j) << std::endl;
// 	}
//       }
//     }
//     for(std::vector<Equation*>::size_type e=0;e<Equation::all().size();e++){
//       const Equation &eqn = *Equation::all()[e];
//       if(eqn.is_active(this)) {
// 	for(int j=0; j<eqn.dim(); j++) {
// 	  std::cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
// 	       << *eqn.nodaleqn(node, j) << std::endl;
// 	}
//       }
//     }
//   }
// }


// void CSubProblem::dump_eqn() const {
//   for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i) {
//     const FuncNode &node = *i.node();
//     std::cout << "node = " << node << std::endl;
//     for(std::vector<Equation*>::size_type e=0;e<Equation::all().size();e++){
//       const Equation &eqn = *Equation::all()[e];
//       if(eqn.is_active(this)) {
// 	for(int j=0; j<eqn.dim(); j++) {
// 	  std::cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
// 	       << *eqn.nodaleqn(node, j) << std::endl;
// 	}
//       }
//     }
//   }
// }

int CSubProblem::check_symmetricity(const CSRmatrix &A) {
  for (int i=0; i<A.nrows(); i++)
    for (int j=0; j<A.ncols(); j++)
      if (A(i,j) != A(j,i)) {
	std::cerr << "(i,	j)= (" << i <<"," << j << ")" << std::endl;
	std::cerr << "A(i,j)=" << A(i,j) << " A(j,i)=" << A(j,i) << std::endl;
	return 0;
      }
  return 1;
}


void CSubProblem::requirePrecompute() {
  precomputeLock.acquire();
  precomputeRequired = true;
  precomputeLock.release();
}


///////////////////////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_MPI

#include "common/mpitools.h"

//This should be called after set_equation_mapping and mapdofeqs.
//A note that might be useful and cleared up my own confusion:
//When an equation (neq=mesh->nodaleqn[i]) is made dependent in bdycondition.py,
//it is the object mesh->nodaleqn[(*global_equation_map)[neq->ndq_index()]]
//that is made dependent, even though it is neq and not the above that
//originally corresponded to the equation. It comes out alright though,
//because when tagging a row (say j) of the local matrix as dependent,
//it suffices to check mesh->nodaleqn[j].
void CSubProblem::set_parallel_mapping()
{
  ResetSharesData();

  // Using global_equation_map, give CNodeShareInfo the information about where
  // in the local stiffness matrix the equation/row can be found.
  for(std::map<NodalEquation*, FuncNode*>::iterator it=mesh->m_eqnnodemap.begin();
      it!=mesh->m_eqnnodemap.end();it++)
    {
      NodalEquation* pne=it->first;
      FuncNode* pfn=it->second;
      //Get sharing information for FuncNode
      CNodeShareInfo* pns=mesh->m_nodesharemap[pfn];
      //Get index of NodalEqn in FuncNode->eqnlist
      int localindex=(find(pfn->eqnlist.begin(),pfn->eqnlist.end(),pne)-
      		      pfn->eqnlist.begin());
      pns->localeqnindexlist.push_back(localindex);
      pns->symmatrixeqnindexlist.push_back((*global_equation_map)[pne->ndq_index()]);
      // ((*global_equation_map)[pne->ndq_index()] is the row number in the stiffness matrix.
      // We shall put this row later into its right place in the combined linear system.
    }
  // Give CNodeShareInfo the information about where in the local
  // stiffness matrix the dof/column can be found
  for(std::map<DegreeOfFreedom*, FuncNode*>::iterator it=mesh->m_dofnodemap.begin();
      it!=mesh->m_dofnodemap.end();it++)
    {
      DegreeOfFreedom* pdof=it->first;
      //FuncNode associated with this DoF
      FuncNode* pfn=it->second;
      //Get sharing information for FuncNode
      CNodeShareInfo* pns=mesh->m_nodesharemap[pfn];
      //Get index of DoF in FuncNode->doflist
      int localindex=(find(pfn->doflist.begin(),pfn->doflist.end(),pdof)-
      		      pfn->doflist.begin());
      pns->localdofindexlist.push_back(localindex);
      //Get index of DoF in the mesh->dof vector
      int mtxindex=(find(mesh->dof.begin(),mesh->dof.end(),pdof)-mesh->dof.begin());
      pns->symmatrixdofindexlist.push_back(mtxindex);
    }
  //RCL testing
  if(Rank()==-1)
    {
      for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=mesh->m_nodesharemap.begin();
	  it!=mesh->m_nodesharemap.end();it++)
	{
	  CNodeShareInfo* pns=it->second;
	  std::cerr << "****** Inherited index ****** " << pns->inheritedindex << std::endl;
	  for(std::vector<int>::size_type i=0;i<pns->localeqnindexlist.size();i++)
	    {
	      std::cerr << "NodalEqn " << pns->localeqnindexlist[i] << "," <<
		pns->symmatrixeqnindexlist[i] << std::endl;
	    }
	  for(std::vector<int>::size_type i=0;i<pns->localdofindexlist.size();i++)
	    {
	      std::cerr << "DoF " << pns->localdofindexlist[i] << "," <<
		pns->symmatrixdofindexlist[i] << std::endl;
	    }
	}
    }

  // This process may own a node but
  // is not associated with an element that is being solved (e.g. the node is at the boundary
  // of a subproblem, and this boundary (zero area) belongs to this process). Have to pass ownership
  // of the node to the next process that can solve for it (i.e. contains an element of the
  // subproblem). (This loop got added after the subproblems were introduced)
  int procSize=Size(),localrank=Rank();
  for(int i=0;i<procSize;i++)// Pass the ownership of nodes to successively higher ranks
    {
      std::vector<int> sbuf;
      // Collect the index of nodes that are not part of the subproblem
      for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=mesh->m_nodesharemap.begin();
	  it!=mesh->m_nodesharemap.end();it++)
	{
	  CNodeShareInfo* pns=it->second;
	  if(pns->owns && !(pns->hasElement))
	    {
	      pns->owns=false;
	      // If everything is consistent, nextrank must be a valid process number (!=-1)
	      std::cout << pns->nextrank << "*** next rank index ***" <<pns-> nextindex << std::endl;
	      sbuf.push_back(pns->nextrank);
	      sbuf.push_back(pns->nextindex);
	    }
	}
      // Gather the sizes to determine the maximum message size from each process
      std::vector<int>* psbufsizelist=_Allgather_Int(sbuf.size());

      std::vector<int>::size_type maxsbufsize=sbuf.size();
      for(std::vector<int>::iterator it=psbufsizelist->begin();it!=psbufsizelist->end();it++)
	{
	  if(maxsbufsize<(std::vector<int>::size_type)*it)
	    maxsbufsize=*it;
	}
      //First element of sbuf must exist, so that _Allgather_IntVec won't crash the program
      if(sbuf.size()==0)
	sbuf.push_back(-1);
      std::vector<int>* psbufall=_Allgather_IntVec(&sbuf,maxsbufsize);
      if(localrank==0)
	{
	  delete psbufsizelist;
	  delete psbufall;
	  continue;
	}
      for(std::vector<int>::size_type k=0;k<psbufall->size();k+=maxsbufsize)
	{
	  int iproc=k/maxsbufsize;
	  for(std::vector<int>::size_type l=0;l<(*psbufsizelist)[iproc];l+=2)
	    {
	      if(localrank==(*psbufall)[k+l])
		{
		  if(mesh->m_indexnodemap.find((*psbufall)[k+l+1])!=mesh->m_indexnodemap.end())
		    {
		      CNodeShareInfo* pns=
			mesh->m_nodesharemap[mesh->m_indexnodemap[(*psbufall)[k+l+1]]];
		      pns->owns=true;
		    }
		  else
		    {
		      //The alternative should not happen
		      std::cerr << "remote index not found!" << std::endl;
		    }
		}
	    }
	}
      delete psbufsizelist;
      delete psbufall;
    }

  //Use information in the vectors indepeqnmap and freedofmap to determine where
  //the rows and columns (DoFs and NodalEqns) should go.
  //If the process owns the node (ergo the DoF and NodalEqn), and if the DoF is free (not fixed),
  //or if the NodalEqn is independent (and active), then the locations should be taken from
  // indepeqnmap or freedofmap.
  //Otherwise, this information should come from the other processes (later).
  m_precombined_indepeqnmap.resize(indepeqnmap.size(),0);
  for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=mesh->m_nodesharemap.begin();
      it!=mesh->m_nodesharemap.end();it++)
    {
      CNodeShareInfo* pns=it->second;
      if(pns->owns==false)
	{
	  for(std::vector<int>::iterator it2=pns->symmatrixeqnindexlist.begin();
	      it2!=pns->symmatrixeqnindexlist.end();it2++)
	    {
	      m_precombined_indepeqnmap[*it2]=-1;//mark as not owned
	    }
	}
    }
  int nrc=0;
  for(std::vector<int>::size_type i=0;i<indepeqnmap.size();i++)
    {
      if(m_precombined_indepeqnmap[i]==0 && indepeqnmap[i]!=-1)
	{
	  m_precombined_indepeqnmap[i]=nrc++;
	}
      else
	{
	  m_precombined_indepeqnmap[i]=-1;//make sure to mark as inactive or not owned
	}
    }
  m_precombined_eqnsize=nrc;

  m_precombined_freedofmap.resize(freedofmap.size(),0);
  for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=mesh->m_nodesharemap.begin();
      it!=mesh->m_nodesharemap.end();it++)
    {
      CNodeShareInfo* pns=it->second;
      if(pns->owns==false)
	{
	  for(std::vector<int>::iterator it2=pns->symmatrixdofindexlist.begin();
	      it2!=pns->symmatrixdofindexlist.end();it2++)
	    {
	      m_precombined_freedofmap[*it2]=-1;//mark as not owned
	    }
	}
    }
  nrc=0;
  for(std::vector<int>::size_type i=0;i<freedofmap.size();i++)
    {
      if(m_precombined_freedofmap[i]==0 && freedofmap[i]!=-1)
	{
	  m_precombined_freedofmap[i]=nrc++;
	}
      else
	{
	  m_precombined_freedofmap[i]=-1;//make sure to mark as fixed or not owned
	}
    }
  m_precombined_dofsize=nrc;

  //ASSERT(m_precombined_dofsize==m_precombined_eqnsize)
  //std::cerr << "m_precombined_eqnsize,dofsize " << m_precombined_eqnsize << " *** =?= *** " <<
  //  m_precombined_dofsize << std::endl;
  //std::cerr << "indepeqnmap,freedofmap sizes " << indepeqnmap.size() << " *** =?= *** " <<
  //  freedofmap.size() << std::endl;

  //Gather all the DoF sizes
  //std::vector<int> *_Allgather_Int(int);//Caller responsible for deletion of return value!
  //Ship and collect
  std::vector<int>* pdofsizelist=_Allgather_Int(m_precombined_dofsize);
  //ASSERT(pdofsizelist->size()==Size()) //Size() from mpitools.h
  if(Rank()==-1)
    {
      for(std::vector<int>::size_type i=0;i<pdofsizelist->size();i++)
	{
	  std::cerr << "***** dofsize ******* " << i << ":" << (*pdofsizelist)[i] << std::endl;
	}
    }

  //Gather the DoF and NodalEqn information owned by each process
  std::vector<int> sbuf;
  for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=mesh->m_nodesharemap.begin();
      it!=mesh->m_nodesharemap.end();it++)
    {
      CNodeShareInfo* pns=it->second;
      if(pns->owns==true)
	{
	  //Move through each process that share the node (but don't own it)
	  //(There usually is only one other process that shares this node)
	  for(std::vector<int>::size_type i=0;i<pns->remoteproclist.size();i++)
	    {
	      std::vector<int>::size_type localdofsize=pns->localdofindexlist.size(),
		localeqnsize=pns->localeqnindexlist.size();

	      //We pack and ship the following:
	      // [remoteproc,remoteindex,#localdofs,#localeqns,
	      //                         localdofindex1,precombineddofindex1,...,
	      //                         localeqnindex1,precombinedeqnindex1,...]
	      //Store the "headers"
	      sbuf.push_back(pns->remoteproclist[i]);
	      sbuf.push_back(pns->remoteindexlist[i]);
	      //Store the dof indices local to the node. It should be the same for every node,
	      // because the fields and equations are added the same way to each node.
	      std::vector<int> sbufdof;
	      //TODO: Optimize later. The fixed DoFs and dependent equations need not have
	      // their sharing information sent.
	      for(std::vector<int>::size_type j=0;j<localdofsize;j++)
		{
		  //Since the node owns it, the map should not return a -1 when
		  // indexed with symmatrixdofindexlist
		  int precomindex=m_precombined_freedofmap[pns->symmatrixdofindexlist[j]];
		  //if(Rank()==0)
		  //  std::cerr << "dof precomindex " << precomindex << "\n";
		  if(precomindex>-1)
		    {
		      sbufdof.push_back(pns->localdofindexlist[j]);
		      sbufdof.push_back(precomindex);
		    }
		}

	      //Store the eqn indices local to the node. It should be the same for every node,
	      // because the fields and equations are added the same way to each node.
	      // (using CNodeShareInfo::doflist and CNodeShareInfo::eqnlist)
	      std::vector<int> sbufeqn;
	      for(std::vector<int>::size_type j=0;j<localeqnsize;j++)
		{
		  int precomindex=m_precombined_indepeqnmap[pns->symmatrixeqnindexlist[j]];
		  //if(Rank()==0)
		  //  std::cerr << "eqn precomindex " << precomindex << "\n";
		  if(precomindex>-1)
		    {
		      sbufeqn.push_back(pns->localeqnindexlist[j]);
		      sbufeqn.push_back(precomindex);
		    }
		}

	      sbuf.push_back(sbufdof.size()/2);
	      sbuf.push_back(sbufeqn.size()/2);
	      for(std::vector<int>::iterator it2=sbufdof.begin();it2!=sbufdof.end();it2++)
		{
		  sbuf.push_back(*it2);
		}
	      for(std::vector<int>::iterator it2=sbufeqn.begin();it2!=sbufeqn.end();it2++)
		{
		  sbuf.push_back(*it2);
		}
	    }
	}
    }

  //Ship and collect
  //First gather the maximum size of sbuf among the processes, then use this in _Allgather_IntVec
  //TODO(?): Explore variable lengths for the buffer size coming from each process ("AllgatherV")
  std::vector<int>* psbufsizelist=_Allgather_Int(sbuf.size());
  std::vector<int>::size_type maxsbufsize=sbuf.size();
  for(std::vector<int>::iterator it=psbufsizelist->begin();it!=psbufsizelist->end();it++)
    {
      if(maxsbufsize<(std::vector<int>::size_type)*it)
	maxsbufsize=*it;
    }

  //First element of sbuf must exist, so that _Allgather_IntVec won't crash the program
  if(sbuf.size()==0)
    sbuf.push_back(-1);
  std::vector<int>* psbufall=_Allgather_IntVec(&sbuf,maxsbufsize);
  //RCL testing, check what's being gathered
  if(Rank()==-1)
    {
      for(std::vector<int>::size_type i=0;i<psbufall->size();i+=maxsbufsize)
	{
	  int iproc=i/maxsbufsize;
	  std::cerr << "\n**** From Proc " << iproc << ", sbufsize=" << (*psbufsizelist)[iproc] <<
	    ", maxsbufsize=" << maxsbufsize << " ****\n";
	  for(std::vector<int>::size_type j=0;j<(*psbufsizelist)[iproc];j++)
	    {
	      std::cerr << (*psbufall)[i+j] << ",";
	    }
	}
    }

  //Unite! Construct mapping for combined stiffness matrix!

  //Check freedofmap and m_precombined_freedofmap before the shift
  if(Rank()==-1)
    {
      for(std::vector<int>::size_type i=0;i<freedofmap.size();i++)
	{
	  std::cerr << "**** freedofmap, m_precombined : " << freedofmap[i] << ", " <<
	    m_precombined_freedofmap[i] << "\n";
	}
      for(std::vector<int>::size_type i=0;i<indepeqnmap.size();i++)
	{
	  std::cerr << "**** indepeqnmap, m_precombined : " << indepeqnmap[i] << ", " <<
	    m_precombined_indepeqnmap[i] << "\n";
	}
    }

  //First calculate offsets for the rows and columns of the local stiffness matrix
  int rowoffset, columnoffset;
  //int localrank=Rank();
  //ASSERT(localrank<pdofsizelist->size())
  rowoffset=0;
  for(std::vector<int>::size_type i=0;i<(std::vector<int>::size_type)localrank;i++)
    {
      rowoffset+=(*pdofsizelist)[i];
    }
  columnoffset=rowoffset;
  //Add all the dofsizes gathered to get the dimension of the global/combined stiffness matrix
  m_combinedmatrixdim=rowoffset;
  for(std::vector<int>::size_type i=(std::vector<int>::size_type)localrank;
      i<pdofsizelist->size();i++)
    {
      m_combinedmatrixdim+=(*pdofsizelist)[i];
    }
  //Shift the indices in m_precombined_freedofmap and m_precombined_indepeqnmap
  for(std::vector<int>::size_type i=0;i<m_precombined_freedofmap.size();i++)
    {
      if(m_precombined_freedofmap[i]!=-1)
	m_precombined_freedofmap[i]+=columnoffset;
    }
  for(std::vector<int>::size_type i=0;i<m_precombined_indepeqnmap.size();i++)
    {
      if(m_precombined_indepeqnmap[i]!=-1)
	m_precombined_indepeqnmap[i]+=rowoffset;
    }

  //Now get the information gathered for the sharing of the DoFs and NodalEqns
  // to shift the indices of the local non-owned rows and columns
  //Note: If the value of indepeqnmap[i] (freedofmap[i]) is non-zero, then
  // m_precombined_indepeqnmap (m_precombined_freedofmap[i]) must also be non-zero.
  rowoffset=0;
  columnoffset=0;
  for(std::vector<int>::size_type i=0;i<psbufall->size();i+=maxsbufsize)
    {
      int iproc=i/maxsbufsize;
      std::vector<int>::size_type realbufsize=(*psbufsizelist)[iproc],
	inc=i;
      if(iproc!=localrank)//No need to check data sent by the same process
	{
	  while(realbufsize>(inc-i))
	    {
	      int data1=(*psbufall)[inc],
		data2;
	      std::vector<int>::size_type tmpdofsize=(*psbufall)[inc+2],
		tmpeqnsize=(*psbufall)[inc+3];
	      if(data1==localrank)
		{
		  //remote index relative to iproc, but local index relative to localrank
		  data1=(*psbufall)[inc+1];
		  CNodeShareInfo* pns=mesh->m_nodesharemap[mesh->m_indexnodemap[data1]];
		  //Go over the localdofindices and the remote dof indices they are mapped to
		  //tmpdofsize=(*psbufall)[inc+2];
		  if(pns->hasElement)
		    {
		      for(std::vector<int>::size_type j=0;j<tmpdofsize;j++)
			{
			  data1=(*psbufall)[inc+4+2*j];
			  data2=(*psbufall)[inc+4+2*j+1];
			  //We expect these find()'s to be successful (i.e. not .end())
			  int localindex=(find(pns->localdofindexlist.begin(),pns->localdofindexlist.end(),data1)-
					  pns->localdofindexlist.begin());
			  //symmatrixdofindexlist[localindex] gives the local target column
			  //when the local stiffness matrix is populated.
			  m_precombined_freedofmap[pns->symmatrixdofindexlist[localindex]]=data2+columnoffset;
			}
		      std::vector<int>::size_type eqnbegin=inc+4+2*tmpdofsize;
		      //tmpeqnsize=(*psbufall)[inc+3];
		      for(std::vector<int>::size_type j=0;j<tmpeqnsize;j++)
			{
			  data1=(*psbufall)[eqnbegin+2*j];
			  data2=(*psbufall)[eqnbegin+2*j+1];
			  //We expect these find()'s to be successful (i.e. not .end())
			  int localindex=(find(pns->localeqnindexlist.begin(),pns->localeqnindexlist.end(),data1)-
					  pns->localeqnindexlist.begin());
			  //symmatrixeqnindexlist[localindex] gives the local target row
			  //when the local stiffness matrix is populated.
			  m_precombined_indepeqnmap[pns->symmatrixeqnindexlist[localindex]]=data2+rowoffset;
			}
		    }
		}
	      inc+=4+2*(tmpdofsize+tmpeqnsize);
	    }
	}
      rowoffset+=(*pdofsizelist)[iproc];
      columnoffset=rowoffset;
    }

  //Check freedofmap and m_precombined_freedofmap after the shift
  if(Rank()==-1)
    {
      for(std::vector<int>::size_type i=0;i<freedofmap.size();i++)
	{
	  std::cerr << "**** freedofmap, m_precombined : " << freedofmap[i] << ", " <<
	    m_precombined_freedofmap[i] << "\n";
	}
      for(std::vector<int>::size_type i=0;i<indepeqnmap.size();i++)
	{
	  std::cerr << "**** indepeqnmap, m_precombined : " << indepeqnmap[i] << ", " <<
	    m_precombined_indepeqnmap[i] << "\n";
	}
    }

  //cleanup
  sbuf.clear();
  psbufsizelist->clear();
  delete psbufsizelist;
  psbufall->clear();
  delete psbufall;
  pdofsizelist->clear();
  delete pdofsizelist;
}

void CSubProblem::NodalPositionSolution(double* temp_array)
{
  //Display (X,Y:NodalValues) at the terminal
  //For testing, and can be called only after set_precombined_equation_mapping and solve
  for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni)
    //for(FuncNodeIterator ni(this); !ni.end(); ++ni)
    {
      FuncNode* pfn=ni.node();
      std::cerr << "(x,y:nodalvalues)=(" << pfn->position()(0) << "," <<
	pfn->position()(1) << ":";
      //Search dof for the index of the dofs in FuncNode->doflist
      for(std::vector<DegreeOfFreedom*>::iterator it=pfn->doflist.begin();
	  it!=pfn->doflist.end();it++)
	{
	  int index=(find(mesh->dof.begin(),mesh->dof.end(),*it)-mesh->dof.begin());
	  if(m_precombined_freedofmap[index]!=-1)
	    std::cerr << temp_array[m_precombined_freedofmap[index]] << ",";
	}
      std::cerr << ")\n";
    }

  // Do work like that of FEMesh::set_unknowns()
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<mesh->dof.size(); i++)
    {
      if(m_precombined_freedofmap[i] != -1)
	{
	  mesh->dof[i]->value() = temp_array[m_precombined_freedofmap[i]];
	}
    }

}

void CSubProblem::ResetSharesData()
{
  //Note: m_precombined_indepeqnmap.resize(...,0) does not reset the elements
  // that have previously been allocated so have to do an explicit clear.
  m_precombined_indepeqnmap.clear();
  m_precombined_freedofmap.clear();
  for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=mesh->m_nodesharemap.begin();
      it!=mesh->m_nodesharemap.end();it++)
    {
      CNodeShareInfo* pns=it->second;
      pns->localdofindexlist.clear();
      pns->localeqnindexlist.clear();
      pns->symmatrixdofindexlist.clear();
      pns->symmatrixeqnindexlist.clear();

      //Reinitialize the owns flag
      if(pns->_owns0)
	pns->owns=true;
      else
	pns->owns=false;

      // Is the node part of an element contained in a subproblem of the current process?
      FuncNode* pfn=it->first;
      if(containsNode(pfn))
	{
	  mesh->m_nodesharemap[pfn]->hasElement=true;
	}
      else
	{
	  mesh->m_nodesharemap[pfn]->hasElement=false;
	}
    }
}

int CSubProblem::GatherNumNodes()
{
  // Get the number of nodes for this subproblem
  // Must be "owned" by the current process.
  int nOwned=0;
  for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni)
    {
      FuncNode* pfn=ni.node();
      if(pfn->isShared())
	{
	  if(mesh->m_nodesharemap[pfn]->_owns0)
	    nOwned++;
	}
      else
	{
	  nOwned++;
	}
    }

  // collect from all processes
  int nnodes=0;
  std::vector<int>* pnnodes=_Allgather_Int(nOwned);
  for(std::vector<int>::size_type i=0;i<pnnodes->size();i++)
    {
      nnodes+=(*pnnodes)[i];
    }
  delete pnnodes;
  return nnodes;
}

#endif

/////////////////////////////////////////////////////////////////////////////

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// ZZ error estimation stuff. Translocated from femesh.C
// TODO: This should be encasulated
// elsewhere, so that different error estimators can be used.

// initializing "scpatches"
void CSubProblem::init_scpatches(const std::vector<int> *snodes) {
  // snodes is a list of mesh node indices, in order of the
  // corresponding skeleton node indices.  scpatches is a std::map,
  // mapping ints to NodalSCPatches*.
  for(std::vector<int>::size_type i=0; i<snodes->size(); i++)
    scpatches[(*snodes)[i]] = new NodalSCPatches();
}

// create/add a new CSCPatch pointer
void CSubProblem::add_scpatch(const int assembly_node, const Material *mat,
			      const int order,
			      const std::vector<int> *elems,
			      const std::vector<int> *nodes,
			      const int qualified)
{
  // make sure to delete new'd CSCPatch's at FEMesh's destruction.
  CSCPatch *patch = new CSCPatch(this, order, mat, elems, nodes, qualified);
  scpatches[assembly_node]->add_patch(patch);
}

// get elements and nodes from the patch
std::vector<int> *CSubProblem::get_elements_from_patch(const int assembly_node,
						       const Material *mat)
{
  return scpatches[assembly_node]->get_elements_from_patch(mat);
}

std::vector<int> *CSubProblem::get_nodes_from_patch(const int assembly_node,
						    const Material *mat)
{
  return scpatches[assembly_node]->get_nodes_from_patch(mat);
}

// initialize "recovered_fluxes"
void CSubProblem::init_nodalfluxes() {
  for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i)
    recovered_fluxes[i.node()->index()] = new NodalFluxes();
}

// recover fluxes
void CSubProblem::recover_fluxes() {
  std::map<const int, NodalSCPatches*>::iterator iter;
  std::vector<Flux*> allfluxes = allFluxes();
  for(iter=scpatches.begin(); iter!=scpatches.end(); iter++)
    (iter->second)->recover_fluxes(allfluxes);
}

// add a recovered flux to the map
void CSubProblem::add_this_flux(const Material *mat, const Flux *fluks,
				const Node *nd, const VECTOR_D *fvalues)
{
  recovered_fluxes[nd->index()]->add_flux_value(mat, fluks, fvalues);
}

// compute a recovered flux at MasterCoord
VECTOR_D *CSubProblem::get_recovered_flux(const Flux *fluks,
					  const Element *elem,
					  const MasterCoord &mc)
{
  int dim = fluks->ndof();
  VECTOR_D *result = new VECTOR_D(dim, 0.0);
  for(int i=0; i<dim; i++) {
    // interpolating the value
    for(ElementFuncNodeIterator node = elem->funcnode_iterator(); !node.end();
	++node)
      {
	double temp = recovered_fluxes[(node.node())->index()]->
	  get_flux_component(elem->material(), fluks, i);
	(*result)(i) += node.shapefunction(mc)*temp;
      }
  }
  return result;
}

// compute all the recovered fluxes at Coord -- swigged one
void CSubProblem::report_recovered_fluxes(const Element *elem,
					  const Coord *point) {
  std::vector<Flux*> fluxes = allFluxes();
  for (std::vector<Flux*>::size_type i=0; i<fluxes.size(); i++) {
    std::cerr << fluxes[i]->name() << std::endl;
    int dim = fluxes[i]->ndof();
    for(int j=0; j<dim; j++) {
      double j_th_component = 0.0;
      for(ElementFuncNodeIterator node = elem->funcnode_iterator(); !node.end();
	  ++node)
	{
	  double comp = recovered_fluxes[(node.node())->index()]
	    ->get_flux_component(elem->material(), fluxes[i], j);
	  j_th_component += node.shapefunction(elem->to_master(*point))*comp;
	}
      std::cerr << j_th_component << std::endl;
    }
    std::cerr << "==========================" << std::endl;
  }
}


double CSubProblem::zz_L2_estimate(const Element *elem, const Flux *fluks) {
  // What to compute:
  // Integral_e[(flux_ex-flux_fe)^T (flux_ex-flux_fe)]^(1/2)
  // flux_ex -> recovered flux
  // 
  // Things to consider to determine integration order
  // Linear mesh: recovered stress field is linear, so the order
  // is 2 -- 2x2 gauss point needed.
  // Quadratic mesh: recovered stress field is quadratic, thus the order
  // should be 4 -- 3x3 gauss points needed.

  int order = 2*elem->shapefun_degree();
  int dim = fluks->ndof();
  double error = 0.0;
  double refer = 0.0;
  for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
    MasterCoord mc = gpt.gausspoint().mastercoord();
    double wt = gpt.gausspoint().weight();
    zz_L2_estimate_sub(elem, fluks, dim, error, refer, mc, wt);
  }
  if (refer == 0.0)
    return 0.0;
  return sqrt(error/refer);
}


void CSubProblem::zz_L2_estimate_sub(const Element *elem, const Flux *fluks,
				     const int &dim, double &error,
				     double &refer,
				     const MasterCoord &mc, const double &wt)
{
  // flux(diff) vectors
  VECTOR_D *recovered = get_recovered_flux(fluks, elem, mc);
  //TODO: Should fluks->evaluate accept a subproblem instead of a
  //mesh object for its first argument?
  VECTOR_D *feflux = fluks->evaluate(this->mesh, elem, mc);
  VECTOR_D diff = *recovered - *feflux;
  double temperror=0.0,temprefer=0.0;
  // dot product while integrating
  for(int i=0; i<dim; i++) {
    //error += diff(i)*diff(i);
    temperror += diff(i)*diff(i);
    //refer += (*recovered)(i) * (*recovered)(i);
    temprefer += (*recovered)(i) * (*recovered)(i);
  }
  //error *= wt;
  error+=temperror*wt;
  //refer *= wt;
  refer+=temprefer*wt;
  delete recovered;
  delete feflux;
}

std::vector<double> *CSubProblem::zz_L2_weights(const Flux *fluks,
						const double &bottom,
						const double &top)
{
  std::vector<double> *weights = new std::vector<double>;
  std::vector<double> values;
  
  int dim = fluks->ndof();
  bool first = true;
  int order=0;			// initialized to suppress compiler warnings
  double min=0, max=0;		// will be reinitialized below
  for(ElementIterator i=element_iterator(); !i.end(); ++i) {
    Element *elem = i.element();
    if(first)
      order = 2*elem->shapefun_degree();
    double value = 0.0;
    for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
      MasterCoord mc = gpt.gausspoint().mastercoord();
      double wt = gpt.gausspoint().weight();
      zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
    }
    values.push_back(value);
    if(first) {
      min = value;
      max = value;
      first = false;
    }
    else {
      if(value < min)
	min = value;
      if(value > max)
	max = value;
    }
  }

//   int order = 2*mesh->element[0]->shapefun_degree();
//   int dim = fluks->ndof();

//   Element *elem = element[0];
//   double value = 0.0;
//   for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
//     MasterCoord mc = gpt.gausspoint().mastercoord();
//     double wt = gpt.gausspoint().weight();
//     zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
//   }
//   values.push_back(value);
//   double min = value;
//   double max = value;

//   for(int i=1; i<mesh->nelements(); i++) {
//     Element *elem = mesh->element[i];
//     value = 0.0;
//     for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
//       MasterCoord mc = gpt.gausspoint().mastercoord();
//       double wt = gpt.gausspoint().weight();
//       zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
//     }
//     values.push_back(value);
//     if(value < min)
//       min = value;
//     if(value > max)
//       max = value;
//   }

  double upper = min + (max-min)*top;
  double lower = min + (max-min)*bottom;
  double diff = upper - lower;
  double sum = upper + lower;
  for(std::vector<double>::size_type i=0; i<values.size(); i++)
    weights->push_back(0.5*erf(M_E*(2*values[i]-sum)/diff)+0.5);
  return weights;
}

void CSubProblem::zz_L2_weights_sub(const Element *elem, const Flux *fluks,
				    const int &dim, double &value,
				    const MasterCoord &mc, const double &wt)
{
  // flux vector
  //TODO: Should evaluate accept a subproblem instead of a mesh object
  //for its first argument?
  VECTOR_D *feflux = fluks->evaluate(this->mesh, elem, mc);
  // dot product while integrating
  double temp=0.0;
  for(int i=0; i<dim; i++) {
    //value += (*feflux)(i)*(*feflux)(i);
    temp += (*feflux)(i)*(*feflux)(i);
  }
  //Seems wrong to multiply wt like this
  //value *= wt;
  value+=temp*wt;
  delete feflux;
}
