// -*- C++ -*-
// $RCSfile: heatconductivity.C,v $
// $Revision: 1.47 $
// $Author: langer $
// $Date: 2008-02-26 22:16:55 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

// heat conductivity

#include <oofconfig.h>
#include "common/coord.h"
#include "common/tostring.h"
#include "engine/csubproblem.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/field.h"
#include "engine/flux.h"
#include "engine/property/heatconductivity/heatconductivity.h"
#include "engine/fieldindex.h"
#include "engine/material.h"
#include "engine/property/orientation/orientation.h"
#include "common/space.h"
#include "engine/nodalequation.h"
#include <iostream>
#include <string>

HeatConductivityProp::HeatConductivityProp(PyObject *reg, const std::string &nm)
  : Property(nm,reg)
{
  temperature = dynamic_cast<ScalarField*>(Field::getField("Temperature"));
  heat_flux = dynamic_cast<VectorFlux*>(Flux::getFlux("Heat_Flux"));
}

int HeatConductivityProp::integration_order(const CSubProblem *subp, 
					    const Element *el) const
{
  if(temperature->in_plane(subp))
    return el->dshapefun_degree();
  return el->shapefun_degree();
}

void HeatConductivityProp::fluxmatrix(const FEMesh *mesh,
				      const Element *el,
				      const ElementFuncNodeIterator &j,
				      const Flux *flux, FluxData *fluxdata,
				      const MasterPosition &pt)
  const
{
  // The heat flux matrix M_{ij} multiplies the vector of nodal
  // temperatures to give the vector heat current J at point pt.
  // M_{ij} = -kappa_{ik} grad_k N_j
  // J_i = -M_{ij} T_j
  // where N_j is the shapefunction at node j.
//   Trace("HeatConductivityProp::fluxmatrix");

  if(*flux != *heat_flux) {
    throw ErrProgrammingError("Unexpected flux", __FILE__, __LINE__);
  }

  double sf = j.shapefunction(pt);
  double dsf0 = j.dshapefunction(0, pt);
  double dsf1 = j.dshapefunction(1, pt);

  const SymmMatrix3 cond(conductivitytensor(mesh, el, pt));

  // Loop over flux components.  Loop over all components, even if
  // the flux is in-plane, because the out-of-plane components of
  // the flux matrix are used to construct the constraint equation.
  for(VectorFieldIterator i; !i.end(); ++i) {
    // in-plane temperature gradient contributions
    fluxdata->matrix_element(i, temperature, j) -=
      cond(i.integer(), 0) * dsf0 + cond(i.integer(), 1) * dsf1;
    
    // out-of-plane temperature gradient contribution
    if(!temperature->in_plane(mesh)) {
      fluxdata->matrix_element(i, temperature->out_of_plane(), j)
	-= cond(i.integer(), 2) * sf;
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

IsoHeatConductivityProp::IsoHeatConductivityProp(PyObject *reg,
						 const std::string &nm,
						 double kppa)
  : HeatConductivityProp(reg, nm),
    kappa_(kppa)
{
}  

void IsoHeatConductivityProp::precompute(FEMesh *mesh) {
  HeatConductivityProp::precompute(mesh);
  conductivitytensor_(0,0) = conductivitytensor_(1,1)
    = conductivitytensor_(2,2) = kappa_;
}

const SymmMatrix3 
IsoHeatConductivityProp::conductivitytensor(const FEMesh*,
					    const Element*,
					    const MasterPosition&)
  const
{
  return conductivitytensor_;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

AnisoHeatConductivityProp::AnisoHeatConductivityProp(PyObject *reg,
						     const std::string &nm,
						     SymmMatrix3 *k)
  : HeatConductivityProp(reg,nm),
    kappa_(*k)
{}

void AnisoHeatConductivityProp::cross_reference(Material *mat) {
  orientation = 
       dynamic_cast<OrientationPropBase*>(mat->fetchProperty("Orientation"));
}

void AnisoHeatConductivityProp::precompute(FEMesh *mesh) {
  HeatConductivityProp::precompute(mesh);
  Trace("AnisoHeatConductivityProp::precompute");
  if(orientation->constant_in_space())
    conductivitytensor_ = kappa_.transform(orientation->orientation());
}

const SymmMatrix3
AnisoHeatConductivityProp::conductivitytensor(const FEMesh *mesh,
					      const Element *el,
					      const MasterPosition &mpos)
  const
{
  if(orientation->constant_in_space())
    return conductivitytensor_;
  return kappa_.transform(orientation->orientation(mesh, el, mpos));
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

HeatSourceProp::HeatSourceProp(PyObject *reg, 
			       const std::string &name, double qd)
  : Property(name,reg),
    qdot_(qd) {
    heat_flux = dynamic_cast<VectorFlux*>(Flux::getFlux("Heat_Flux"));
}

int HeatSourceProp::integration_order(const CSubProblem*, const Element*)
  const
{
  return 0;
}

// HeatSource adds to the right-hand side of the heat equation, 
// so we want fluxrhs and not fluxmatrix.
void HeatSourceProp::fluxrhs(const FEMesh *mesh, const Element *element,
			     const Flux *flux, FluxData *fluxdata,
			     const MasterPosition &masterpos) const {

  if(*flux != *heat_flux) {
    throw ErrProgrammingError("Unexpected flux.", __FILE__, __LINE__);
  }

  // In fact, there is only one component, so this is excessive.
//   for(int i=0;i<heat_flux->divergence_dim();i++) {
//     fluxdata->rhs_element(i) -= qdot_;
//   }
  fluxdata->rhs_element(0) -= qdot_;
}
